import matplotlib.pyplot as plt
import numpy as np
from numpy import pi
from scipy.interpolate import Rbf


def myplot(i, x, y, U):
    plt.figure(i)
    ax = plt.gca()

    ax.scatter(x, y, c=U, vmin=-1, vmax=1)


def my_multiplot(nf, x, y, U, n, func):
    plt.figure(nf)
    for i in range(n):
        ax = plt.subplot(1, n, i + 1)
        ax.scatter(x, y, c=U[:, i], vmin=-1, vmax=1)
        ax.set_title(func[i])


# set of random x coordinates from 0 to 2pi
x = 2 * np.pi * np.random.rand(100)
y = 2 * np.pi * np.random.rand(100)

U = np.exp(-((x - pi) ** 2 + (y - pi) ** 2))  # gaussian
# U = x  # linear

myplot(0, x, y, U)


# calculate tps coeff
centers = np.vstack([x, y])
smoothing_coef = 0
subdom_size = 20


# interpolation grid

xI = yI = np.arange(0, 2 * pi, 0.1)
XI, YI = np.meshgrid(xI, yI)
XI = XI.ravel()
YI = YI.ravel()

U_vrai = np.exp(-((XI - pi) ** 2 + (YI - pi) ** 2))

func = (
    "linear",
    "inverse",
    "gaussian",
    "multiquadric",
    "cubic",
    "quintic",
    "thin-plate",
)
n = len(func)
ERR_eval = np.zeros((np.size(XI), n))
for i in range(n):
    print(i)
    rbfi = Rbf(x, y, U, function=func[i])
    ERR_eval[:, i] = (rbfi(XI, YI) - U_vrai) / U_vrai

my_multiplot(1, XI, YI, ERR_eval, n, func)

plt.show()



import shutil
import unittest
from pathlib import Path
from tempfile import gettempdir

import matplotlib

matplotlib.use("Agg")

from fluiddyn.io import stdout_redirected
from fluidimage import get_path_image_samples
from fluidimage.data_objects.tomo import ArrayTomoCV
from fluidimage.reconstruct.tomo import TomoMLOSCV


path_calib = str(get_path_image_samples() / "TomoPIV" / "calibration" / "cam0.h5")
path_particle = str(
    get_path_image_samples()
    / "TomoPIV"
    / "particle"
    / "cam0.pre"
    / "im00001a.tif"
)
path_output = Path(gettempdir()) / "fluidimage_test_mlos"


class TestMLOS(unittest.TestCase):
    """Test fluidimage.reconstruct.tomo.mlos module."""

    def tearDown(self):
        shutil.rmtree(path_output)

    def test(self):
        """Test classes TomoMLOSCV and ArrayTomoCV."""
        with stdout_redirected():
            tomo = TomoMLOSCV(
                path_calib,
                xlims=(-10, 10),
                ylims=(-10, 10),
                zlims=(-5, 5),
                nb_voxels=(11, 11, 5),
            )
            tomo.verify_projection()
            pix = tomo.phys2pix("cam0")
            tomo.array.init_paths(path_particle, path_output)
            tomo.reconstruct(pix, path_particle, threshold=None, save=True)

            path_result = list(path_output.glob("*"))[0]
            array = ArrayTomoCV(h5file_path=path_result)
            array.describe()
            array.load_dataset(copy=True)
            array.plot_slices(0, 1)
            array.clear()


if __name__ == "__main__":
    unittest.main()



"""Post-processing of vector fields (:mod:`fluidimage.postproc.vector_field`)
=============================================================================

.. autoclass:: VectorFieldOnGrid
   :members:
   :private-members:

.. autoclass:: ArrayOfVectorFieldsOnGrid
   :members:
   :private-members:

"""

from copy import deepcopy
from numbers import Number
import itertools
from typing import Optional

import matplotlib.pyplot as plt
import numpy as np
import h5py
from scipy.ndimage import gaussian_filter, median_filter

from fluiddyn.util.paramcontainer import ParamContainer

from .util import (
    reshape_on_grid_final,
    compute_rot,
    compute_div,
    compute_1dspectrum,
    compute_2dspectrum,
)


def _is_regular(x):
    dx = x[1] - x[0]
    return np.allclose(np.diff(x), dx)


class VectorFieldOnGrid:
    """PIV field on a regular grid.

    Parameters
    ----------

    x : np.array (1d)

    y : np.array (1d)

    z : number or np.array (1d)

    vx : np.array (2d)

    vy : np.array (2d)

    vz : np.array (2d), optional

    namevx : str, 'vx'
    namevy : str, 'vy'
    namevz : str, 'vz'

    unitvx : str, '?'
    unitvy : str, '?'
    unitvz : str, '?'

    namex : str, 'x'
    namey : str, 'y'
    namez : str, 'z'

    unitx : str, '?'
    unity : str, '?'
    unitz : str, '?'

    """

    _attr_saved_as_dataset = list(
        "".join(t) for t in itertools.product(["v", ""], "xyz")
    )
    _attr_saved_as_attr = list(
        "".join(t)
        for t in itertools.product(["name", "unit"], _attr_saved_as_dataset)
    ) + ["name", "history"]

    def __init__(
        self,
        x,
        y,
        z,
        vx,
        vy,
        vz=np.nan,
        namevx="vx",
        namevy="vy",
        namevz="vz",
        unitvx="?",
        unitvy="?",
        unitvz="?",
        namex="x",
        namey="y",
        namez="z",
        unitx="?",
        unity="?",
        unitz="?",
        name="Fluidimage_field",
        history=["fluidimage"],
        params: Optional[ParamContainer] = None,
    ):

        if isinstance(z, np.ndarray) and z.ndim == 0:
            z = z.item()

        self.x = x
        self.y = y
        self.z = z
        self.vx = vx
        self.vy = vy
        self.vz = vz
        self.unitvx = unitvx
        self.namevx = namevx
        self.unitvy = unitvy
        self.namevy = namevy
        self.unitvz = unitvz
        self.namevz = namevz
        self.unitx = unitx
        self.namex = namex
        self.unity = unity
        self.namey = namey
        self.unitz = unitz
        self.namez = namez

        self.name = name
        self.history = history
        self.params = params

        self.is_grid_regular = _is_regular(x) and _is_regular(y)

        if z is not None and self.is_grid_regular:
            self.is_grid_regular = (
                isinstance(z, Number) or z.shape == vx.shape or _is_regular(z)
            )

        # data orientation for proper quiver
        flipx = self.x[1] - self.x[0] < 0
        flipy = self.y[1] - self.y[0] < 0
        if flipx:
            self.x = np.flip(self.x)
            self.vx = np.fliplr(self.vx)
            self.vy = np.fliplr(self.vy)
            if not (isinstance(self.vz, Number) or not vz.shape == vx.shape):
                self.vz = np.fliplr(self.vz)
            if z is not None and not (
                isinstance(self.z, Number) or not self.vx.shape == self.z.shape
            ):
                self.z = np.fliplr(self.z)
        if flipy:
            self.y = np.flip(self.y)
            self.vx = np.flipud(self.vx)
            self.vy = np.flipud(self.vy)
            if not (isinstance(self.vz, Number) or not vz.shape == vx.shape):
                self.vz = np.flipud(self.vz)
            if z is not None and not (
                isinstance(self.z, Number) or not self.vx.shape == self.z.shape
            ):
                self.z = np.flipud(self.z)

        if self.is_grid_regular:
            self.dx = self.x[1] - self.x[0]
            self.dy = self.y[1] - self.y[0]

    @property
    def shape(self):
        """Shape of the field (``(nz, ny, nx)`` or ``(ny, nx)``)"""
        if self.z is None or isinstance(self.z, Number) or self.z.ndim > 1:
            return len(self.y), len(self.x)
        else:
            return len(self.z), len(self.y), len(self.x)

    def save(self, path, params=None):
        """Save the object in a hdf5 file"""
        with h5py.File(path, "w") as file:
            cls = self.__class__
            file.attrs["class_name"] = cls.__name__
            file.attrs["module_name"] = cls.__module__

            for attr_name in self._attr_saved_as_dataset:
                file.create_dataset(attr_name, data=getattr(self, attr_name))

            for attr_name in self._attr_saved_as_attr:
                file.attrs[attr_name] = getattr(self, attr_name)

            if params is None and self.params is not None:
                params = self.params

            if params is not None:
                params._save_as_hdf5(hdf5_parent=file)

    @classmethod
    def from_file(cls, path, load_params=False):
        """Create a PIV2d object from a file

        It can be a file representing a LightPIVResults or a PIV2d object.

        """

        with h5py.File(path, "r") as file:
            class_name = file.attrs["class_name"]
            module_name = file.attrs["module_name"]

        if isinstance(class_name, bytes):
            class_name = class_name.decode()
            module_name = module_name.decode()

        if (
            class_name in ("MultipassPIVResults", "LightPIVResults")
            and module_name == "fluidimage.data_objects.piv"
        ):
            with h5py.File(path, "r") as file:
                params = ParamContainer(hdf5_object=file["params"])

            if class_name == "MultipassPIVResults":
                key_piv = f"/piv{params.multipass.number-1}"
            else:
                key_piv = "piv"

            with h5py.File(path, "r") as file:
                piv = file[key_piv]
                ixvecs_final = piv["ixvecs_final"][...]
                iyvecs_final = piv["iyvecs_final"][...]
                deltaxs = piv["deltaxs_final"][...]
                deltays = piv["deltays_final"][...]

                kwargs = {}
                (X, Y, kwargs["vx"], kwargs["vy"],) = reshape_on_grid_final(
                    ixvecs_final, iyvecs_final, deltaxs, deltays
                )

                kwargs["x"] = X[0, :]
                kwargs["y"] = Y[:, 0]

                kwargs["z"] = None
                kwargs["params"] = params

        elif class_name == cls.__name__ and module_name == cls.__module__:
            kwargs = {}
            with h5py.File(path, "r") as file:
                for attr_name in cls._attr_saved_as_dataset:
                    kwargs[attr_name] = file[attr_name][...]
                for attr_name in cls._attr_saved_as_attr:
                    attr = file.attrs[attr_name]
                    if isinstance(attr, np.ndarray):
                        attr = list(attr)
                    kwargs[attr_name] = attr

                if "params" in file.keys():
                    params = ParamContainer(hdf5_object=file["params"])
                    kwargs["params"] = params

        else:
            print(class_name, module_name)
            raise NotImplementedError
        return cls(**kwargs)

    def __add__(self, other):
        if isinstance(other, Number):
            vx = self.vx + other
            vy = self.vy + other
            vz = self.vz + other
        else:
            vx = self.vx + other.vx
            vy = self.vy + other.vy
            vz = self.vz + other.vz

        return type(self)(self.x, self.y, self.z, vx, vy, vz)

    def __radd__(self, other):
        if other == 0:
            return self

        else:
            return self.__add__(other)

    def __sub__(self, other):
        if isinstance(other, Number):
            vx = self.vx - other
            vy = self.vy - other
            vz = self.vz - other
        else:
            vx = self.vx - other.vx
            vy = self.vy - other.vy
            vz = self.vz - other.vz

        return type(self)(self.x, self.y, self.z, vx, vy, vz)

    def __mul__(self, other):
        if isinstance(other, Number):
            vx = other * self.vx
            vy = other * self.vy
            vz = other * self.vz
        else:
            vx = other.vx * self.vx
            vy = other.vy * self.vy
            vz = other.vz * self.vz
        return type(self)(self.x, self.y, self.z, vx, vy, vz)

    def __rmul__(self, other):
        return self.__mul__(other)

    def __div__(self, other):
        if isinstance(other, Number):
            vx = self.vx / other
            vy = self.vy / other
            vz = self.vz / other
        else:
            vx = self.vx / other.vx
            vy = self.vy / other.vy
            vz = self.vz / other.vz
        return type(self)(self.x, self.y, self.z, vx, vy, vz)

    def __truediv__(self, other):
        return self.__div__(other)

    def display(
        self, scale=1, background=None, ax=None, skip=(slice(None), slice(None))
    ):
        """Display the vector field"""

        if background is not None:
            raise NotImplementedError

        if ax is None:
            fig, ax = plt.subplots()
        else:
            fig = ax.figure

        pcm = ax.pcolormesh(self.x, self.y, self.compute_norm())

        # minus because the y axis is inverted
        q = ax.quiver(
            self.x[skip[0]],
            self.y[skip[1]],
            self.vx[skip],
            self.vy[skip],
            scale_units="xy",
            scale=scale,
        )
        ax.set_xlabel(self.namex + " [" + self.unitx + "]")
        ax.set_ylabel(self.namey + " [" + self.unity + "]")

        def onclick(event):
            key = event.key
            if key == "ctrl++":
                q.scale *= 2.0
                print(key + ": multiply q.scale by 2.", end="")
            elif key == "ctrl+-":
                q.scale /= 2.0
                print(key + ": divide q.scale by 2.", end="")
            if event.key in ["ctrl++", "ctrl+-"]:
                print(f" q.scale = {q.scale}")
                fig.canvas.draw()

        fig.canvas.mpl_connect("key_press_event", onclick)
        plt.colorbar(pcm, ax=ax)

        xmin = self.x.min()
        xmax = self.x.max()
        lx = xmax - xmin
        ymin = self.y.min()
        ymax = self.y.max()
        ly = ymax - ymin
        n = 20
        ax.set_xlim([xmin - lx / n, xmax + lx / n])
        # y axis inverted!
        # ax.set_ylim([ymax + ly / n, ymin - ly / n])
        ax.set_ylim([ymin - ly / n, ymax + ly / n])
        ax.set_aspect("equal")

        return ax

    def median_filter(self, size, niter=1, valid=True):
        """Return a new field filtered with a median filter"""
        if np.isscalar(size):
            size = [size, size]

        def _medianf(f):
            for i in range(niter):
                f = median_filter(f, size)
            return f

        ret = deepcopy(self)
        ret.vx = _medianf(self.vx)
        ret.vy = _medianf(self.vy)

        if hasattr(self, "vz"):
            if isinstance(self.vz, Number):
                ret.vz = self.vz
            else:
                ret.vz = _medianf(self.vz)

        if valid:
            mf = int(np.floor(max(size) / 2))
            ny, nx = self.vx.shape
            ret = ret.extract(mf, ny - mf, mf, nx - mf)

        ret.history.append(
            f"median_filter(size={size}, niter={niter}, valid ={valid})"
        )

        return ret

    def gaussian_filter(self, sigma, niter=1, truncate=3, valid=True):
        """Return a new field filtered with a gaussian filter"""
        if np.isscalar(sigma):
            sigma = [sigma, sigma]

        def _gaussianf(f):
            for i in range(niter):
                f = gaussian_filter(f, sigma, truncate=truncate)
            return f

        ret = deepcopy(self)
        ret.vx = _gaussianf(self.vx)
        ret.vy = _gaussianf(self.vy)

        if hasattr(self, "vz"):
            # ret.vz = _gaussianf(self.vz)
            pass

        if valid:
            mf = int(np.floor((2 * int(truncate * max(sigma) + 0.5) + 1) / 2))
            ny, nx = self.vx.shape
            ret = ret.extract(mf, ny - mf, mf, nx - mf)

        ret.history.append(
            "gaussian_filter(sigma={}, niter={}, valid={})".format(
                sigma, niter, valid
            )
        )

        return ret

    def extract(self, start0, stop0, start1, stop1, phys=False):
        """Return a new field extrated from `self`"""

        ret = deepcopy(self)

        if phys:
            indy = (ret.y >= start0) & (ret.y <= stop0)
            indx = (ret.x >= start1) & (ret.x <= stop1)
            start0 = np.argwhere(ret.y == ret.y[indy].min())[0][0]
            stop0 = np.argwhere(ret.y == ret.y[indy].max())[0][0] + 1
            start1 = np.argwhere(ret.x == ret.x[indx].min())[0][0]
            stop1 = np.argwhere(ret.x == ret.x[indx].max())[0][0] + 1

        def _extract2d(f):
            return f[start0:stop0, start1:stop1]

        ret.x = ret.x[start1:stop1]
        ret.y = ret.y[start0:stop0]
        ret.vx = _extract2d(ret.vx)
        ret.vy = _extract2d(ret.vy)
        if hasattr(self, "vz") and np.size(self.vz) > 1:
            ret.vz = _extract2d(ret.vz)
        if (
            self.z is not None
            and not isinstance(self.z, Number)
            and self.vx.shape == self.z.shape
        ):
            ret.z = _extract2d(ret.z)

        ret.history.append(
            (
                "extract(start0={}, stop0={}, " "start1={}, stop1={}, phys={})"
            ).format(start0, stop0, start1, stop1, phys)
        )

        return ret

    def truncate(self, cut=1, phys=False):
        """Return a new truncated field"""

        if phys:
            raise NotImplementedError

        ny, nx = self.vx.shape
        return self.extract(cut, ny - cut, cut, nx - cut)

    def extract_square(self, cut=0, force_even=True):
        """Return a square field"""
        n1 = self.x.size
        n0 = self.y.size
        n = min(n0, n1) - 2 * cut

        if force_even and n % 2 == 1:
            n -= 1

        if n1 > n0:
            start0 = cut
            stop0 = cut + n
            start1 = (n1 - n) // 2
            stop1 = start1 + n
        else:
            start1 = cut
            stop1 = cut + n
            start0 = (n0 - n) // 2
            stop0 = start1 + n

        return self.extract(start0, stop0, start1, stop1)

    def compute_norm(self):
        """Compute the norm of the vector field"""
        return np.sqrt(self.vx**2 + self.vy**2)

    def compute_spatial_fft(self, axes=(0, 1)):
        """Compute the spatial Fourier transform"""
        if axes == (0, 1):
            vx_fft, kx, ky, psd_vx = compute_2dspectrum(
                self.x, self.y, self.vx, axes=axes
            )
            vy_fft, kx, ky, psd_vy = compute_2dspectrum(
                self.x, self.y, self.vy, axes=axes
            )
            return vx_fft, vy_fft, kx, ky, psd_vx, psd_vy
        elif axes == 0:
            vx_fft, ky, psd_vx = compute_1dspectrum(self.y, self.vx, axis=axes)
            vy_fft, ky, psd_vy = compute_1dspectrum(self.y, self.vy, axis=axes)
            return vx_fft, vy_fft, self.x, ky, psd_vx, psd_vy
        elif axes == 1:
            vx_fft, kx, psd_vx = compute_1dspectrum(self.x, self.vx, axis=axes)
            vy_fft, kx, psd_vy = compute_1dspectrum(self.x, self.vy, axis=axes)
            return vx_fft, vy_fft, kx, self.y, psd_vx, psd_vy

    def compute_rotz(self, edge_order=2):
        """Compute the vertical curl"""

        if not self.is_grid_regular:
            raise NotImplementedError

        dvx_dy = np.gradient(self.vx, self.dy, axis=0, edge_order=edge_order)
        dvy_dx = np.gradient(self.vy, self.dx, axis=1, edge_order=edge_order)
        return compute_rot(dvx_dy, dvy_dx)

    def compute_divh(self, edge_order=2):
        """Compute the horizontal divergence"""

        if not self.is_grid_regular:
            raise NotImplementedError

        dvx_dx = np.gradient(self.vx, self.dx, axis=1, edge_order=edge_order)
        dvy_dy = np.gradient(self.vy, self.dy, axis=0, edge_order=edge_order)
        return compute_div(dvx_dx, dvy_dy)


class ArrayOfVectorFieldsOnGrid:
    def __init__(self, fields=None):
        if fields is None:
            fields = []
        elif not isinstance(fields, (list, tuple)):
            raise TypeError

        self._list = list(fields)

        self.timestep = None
        self.times = None
        self.unit_time = None

    def set_timestep(self, timestep, unit_time="sec"):
        """Set the timestep (and a time vector)"""
        self.timestep = timestep
        self.unit_time = unit_time
        nb_files = len(self)
        self.times = np.linspace(0, timestep * (nb_files - 1), nb_files)

    def compute_time_average(self):
        """Compute the time average"""
        result = deepcopy(self._list[0])
        for piv in self._list[1:]:
            result += piv
        return result / len(self._list)

    def append(self, v):
        """Append an element"""
        self._list.append(v)

    def extend(self, l):
        """Extend from an iterable"""
        self._list.extend(l)

    def __add__(self, other):
        result = type(self)()
        for v in self._list:
            result.append(v + other)
        return result

    def __radd__(self, other):
        if other == 0:
            return self
        else:
            return self.__add__(other)

    def __sub__(self, other):
        result = type(self)()
        for v in self._list:
            result.append(v - other)
        return result

    def __mul__(self, other):
        result = type(self)()
        for v in self._list:
            result.append(v * other)
        return result

    def __rmul__(self, other):
        result = type(self)()
        for v in self._list:
            result.append(other * v)
        return result

    def __truediv__(self, other):
        result = type(self)()
        for v in self._list:
            result.append(v / other)
        return result

    def __iter__(self):
        return iter(self._list)

    def __getitem__(self, key):
        r = self._list.__getitem__(key)
        if isinstance(r, list):
            r = type(self)(r)
        return r

    def __setitem__(self, key, value):
        self._list.__setitem__(key, value)

    def __delitem__(self, key):
        self._list.__delitem__(key)

    def __repr__(self):
        return (
            "ArrayPIV containing {} fields:\n".format(len(self))
            + self._list.__repr__()
        )

    def __len__(self):
        return self._list.__len__()

    def median_filter(self, size, niter=1, valid=True):
        """Return a new array filtered with a median filter"""
        result = type(self)()
        for v in self:
            result.append(v.median_filter(size, niter=niter, valid=valid))
        return result

    def gaussian_filter(self, sigma, niter=1, truncate=3, valid=True):
        """Return a new array filtered with a gaussian filter"""
        result = type(self)()
        for v in self:
            result.append(
                v.gaussian_filter(
                    sigma, niter=niter, truncate=truncate, valid=valid
                )
            )
        return result

    def extract(self, start0, stop0, start1, stop1, phys=False):
        """Extract new fields from the fields"""
        result = type(self)()
        for v in self:
            result.append(v.extract(start0, stop0, start1, stop1, phys=phys))
        return result

    def truncate(self, cut=1, phys=False):
        """Truncate the fields"""
        result = type(self)()
        for v in self:
            result.append(v.truncate(cut=cut, phys=phys))
        return result

    def extract_square(self, cut=0):
        """Extract square fields"""
        result = type(self)()
        for v in self:
            result.append(v.extract_square(cut=cut))
        return result

    def compute_temporal_fft(self):
        """Compute the temporal Fourier transform"""
        if self.times is None:
            raise RuntimeError(
                "please use `set_timestep` to define time before performing "
                "temporal Fourier transform"
            )

        piv0 = self._list[0]
        fields = np.empty([len(self._list), *piv0.shape])
        for it, piv in enumerate(self._list):
            fields[it, :, :] = piv.vx

        vx_fft, omega, psdU = compute_1dspectrum(self.times, fields, axis=0)

        for it, piv in enumerate(self._list):
            fields[it, :, :] = piv.vy

        vy_fft, omega, psdV = compute_1dspectrum(self.times, fields, axis=0)

        return vx_fft, vy_fft, omega, psdU, psdV

    def apply_function_to_spatiotemp_data(self, func):
        """
        Apply the function func on 3D data vx(t, y, x) and vy(t, y, x).
        """
        piv0 = self._list[0]
        fields = np.empty([len(self._list), *piv0.shape])
        for it, piv in enumerate(self._list):
            fields[it, :, :] = piv.vx
        retvx = func(fields)
        for it, piv in enumerate(self._list):
            fields[it, :, :] = piv.vy
        retvy = func(fields)
        result = deepcopy(self)
        for it, v in enumerate(result):
            v.vx = retvx[it]
            v.vy = retvy[it]
        return result

    def display(
        self,
        ind=0,
        scale=1,
        background=None,
        ax=None,
        skip=(slice(None), slice(None)),
    ):
        ax = self._list[ind].display(scale, background, ax, skip)
        fig = ax.figure
        self.currentind = ind

        def onscroll(event):
            if event.button == "up":
                self.currentind = (self.currentind + 1) % len(self)
            else:
                self.currentind = (self.currentind - 1) % len(self)
            C = self._list[self.currentind].compute_norm()
            ax.collections[0].set_array(C.ravel())
            ax.collections[1].set_UVC(
                self._list[self.currentind].vx[skip],
                self._list[self.currentind].vy[skip],
            )
            print(f"t ={self.times[self.currentind]:.2f} " + self.unit_time)
            fig.canvas.draw()

        fig.canvas.mpl_connect("scroll_event", onscroll)
        return ax



"""Reconstruction algorithms
============================

In general reconstruction simply means applying the calibration of mutiple
cameras to obtain a sense of depth. It is performed prior to PIV in the case of
tomographic PIV (3D-3C PIV) and after PIV in the case of stereo PIV (2D-3C
PIV).

Provides:

.. autosummary::
   :toctree:

   tomo

"""



"""Uvmat interface (:mod:`fluidimage.run_from_xml`)
===================================================

From matlab, something like::

  system(['python -m fluidimage.run_from_xml ' path_to_instructions_xml])


.. todo::

   UVmat civserie: input and output.

"""

import argparse
import os
from time import time

import numpy as np
import scipy

from fluiddyn.util import import_class
from fluiddyn.util.paramcontainer import tidy_container
from fluidimage.topologies.piv import TopologyPIV

from . import (
    ParamContainer,
    SerieOfArraysFromFiles,
    config_logging,
    logger,
    reset_logger,
)


def tidy_uvmat_instructions(params):
    params._set_internal_attr("_tag", "instructions_uvmat")

    # get nicer names and a simpler organization...
    tidy_container(params)

    params.input_table = input_table = params.input_table.split(" & ")
    for i in range(1, len(input_table)):
        if input_table[i].startswith("/"):
            input_table[i] = input_table[i][1:]

    filename = "".join(input_table[2:])
    path_dir = os.path.join(*input_table[:2])
    path_file_input = os.path.abspath(os.path.join(path_dir, filename))
    path_dir_input = path_dir
    params._set_attrib("path_dir_input", path_dir_input)
    params._set_attrib("path_file_input", path_file_input)

    params._set_attrib("path_dir_output", path_dir_input + params.output_dir_ext)

    ir = params.index_range
    if not hasattr(ir, "incr_i"):
        ir._set_attrib("incr_i", 1)

    if not hasattr(ir, "first_i"):
        print("Warning: no attribute first_i in xml UVmat file.")
        ir._set_attrib("first_i", 1)

    if not hasattr(ir, "last_i"):
        print("Warning: no attribute last_i in xml UVmat file.")
        ir._set_attrib("last_i", None)

    if not hasattr(ir, "first_j"):
        strcouple = f"i:i+{ir.incr_i + 1}:{ir.incr_i}"
        ind_start = ir.first_i

        if ir.last_i is None:
            ind_stop = None
        else:
            ind_stop = ir.last_i - ir.incr_i + 1
    else:
        raise NotImplementedError

    params._set_attrib("strcouple", strcouple)
    params._set_attrib("ind_stop", ind_stop)
    params._set_attrib("ind_start", ind_start)


class ActionBase:
    def __init__(self, instructions, args):
        self.instructions = instructions

        # create the serie of arrays
        logger.info("Create the serie of arrays")
        self.serie_arrays = SerieOfArraysFromFiles(
            path=instructions.path_dir_input,
            index_slices=instructions.index_slices,
        )


class ActionAverage(ActionBase):
    """Compute the average and save as a png file."""

    def compute(self):
        instructions = self.instructions
        serie = self.serie_arrays
        # compute the average
        logger.info("Compute the average")
        a = serie.get_array_from_index(0)
        mean = np.zeros_like(a, dtype=np.float32)
        nb_fields = 0
        for a in serie.iter_arrays():
            mean += a
            nb_fields += 1
        mean /= nb_fields

        strindices_first_file = serie._compute_strindices_from_indices(
            *[indices[0] for indices in instructions.index_slices]
        )
        strindices_last_file = serie._compute_strindices_from_indices(
            *[indices[1] - 1 for indices in instructions.index_slices]
        )

        name_file = (
            serie.base_name
            + serie._separator_base_index
            + strindices_first_file
            + "-"
            + strindices_last_file
            + "."
            + serie.extension_file
        )

        path_save = os.path.join(instructions.path_dir_output, name_file)
        logger.info("Save in file:\n%s", path_save)
        scipy.misc.imsave(path_save, mean)
        return mean


def params_piv_from_uvmat_xml(instructions, args):
    params = TopologyPIV.create_default_params()

    params.series.path = instructions.path_file_input

    params.series.strcouple = instructions.strcouple
    params.series.ind_stop = instructions.ind_stop
    params.series.ind_start = instructions.ind_start

    params.saving.path = instructions.path_dir_output

    n0 = int(instructions.action_input.civ1.search_box_size.split("\t")[0])
    if n0 % 2 == 1:
        n0 -= 1

    params.piv0.shape_crop_im0 = n0

    if hasattr(instructions.action_input, "patch2"):
        params.multipass.subdom_size = (
            instructions.action_input.patch2.sub_domain_size
        )
    else:
        params.multipass.use_tps = False

    if hasattr(instructions.action_input, "civ2"):
        params.multipass.number = 2

    modif_fluidimage_params(params, args)

    return params


class ActionPIV(ActionBase):
    """Compute the average and save as a png file."""

    def __init__(self, instructions, args):
        self.instructions = instructions
        self.params = params_piv_from_uvmat_xml(instructions, args)
        logger.info("Initialize fluidimage computations with parameters:")
        logger.info(self.params._make_xml_text())
        self.topology = TopologyPIV(self.params)

    def compute(self):
        t = time()
        self.topology.compute(sequential=False)
        t = time() - t
        print(f"ellapsed time: {t}s")


actions_classes = {"aver_stat": ActionAverage, "civ_series": ActionPIV}

programs = ["uvmat", "fluidimage"]


def parse_args():
    parser = argparse.ArgumentParser(
        description=__doc__, formatter_class=argparse.RawDescriptionHelpFormatter
    )
    parser.add_argument("path", help="Path file.", type=str, nargs="?")
    parser.add_argument(
        "-m",
        "--mode",
        help="‘ask’, ‘new_dir’, ‘complete’ or ‘recompute’.",
        type=str,
        default="ask",
    )

    return parser.parse_args()


def modif_fluidimage_params(params, args):
    try:
        params.saving.how = args.mode
    except AttributeError:
        pass


def main():

    reset_logger()
    config_logging("info")

    args = parse_args()

    path_instructions_xml = args.path
    logger.info(
        "\nFrom Python, start with instructions in xml file:\n%s",
        path_instructions_xml,
    )

    params = ParamContainer(path_file=path_instructions_xml)

    program = None

    try:
        params.Action.ActionName
    except AttributeError:
        pass
    else:
        program = "uvmat"

    try:
        program = params._value_text["program"]
    except (AttributeError, KeyError):
        pass

    if program not in programs:
        raise ValueError("Can not detect the program to launch.")

    print(args)

    print(f"using input file from {program}")

    if program == "uvmat":
        tidy_uvmat_instructions(params)
        action_name = params.action.action_name
        logger.info(
            'Check if the action "%s" is implemented by FluidImage', action_name
        )
        if action_name not in actions_classes.keys():
            raise NotImplementedError(
                'action "' + action_name + '" is not yet implemented.'
            )

        cls = actions_classes[action_name]
        action = cls(params, args)

    elif program == "fluidimage":
        cls = import_class(
            params._value_text["module"], params._value_text["class"]
        )
        modif_fluidimage_params(params, args)
        action = cls(params)

    return action.compute()


if __name__ == "__main__":

    result = main()



import os
import time
from warnings import warn

import numpy as np
import pylab

from fluidimage.data_objects.piv import LightPIVResults

from .util import (
    compute_div,
    compute_rot,
    compute_1dspectrum,
    compute_2dspectrum,
)

warn(
    "Functions and classes defined in this module (fluidimage.postproc.postproc)"
    "are deprecated (the code is not tested!). "
    "Please use fluidimage.postproc.vector_field instead.",
    DeprecationWarning,
)


def compute_ken(U, V):
    ken = (U**2 + V**2) / 2
    return ken


def compute_norm(U, V):
    norm = np.sqrt(U**2 + V**2)
    return norm


def compute_grid(xs, ys, deltaxs, deltays):
    x = np.unique(xs)
    y = np.unique(ys)
    X, Y = np.meshgrid(x, y)
    U = np.reshape(deltaxs, X.shape)
    V = np.reshape(deltays, X.shape)
    return X, Y, U, V


def compute_derivatives(dx, dy, U, V, edge_order=2):
    dUdx, dUdy = np.gradient(U, dx, dy, edge_order=edge_order)
    dVdx, dVdy = np.gradient(V, dx, dy, edge_order=edge_order)
    return dUdx, dUdy, dVdx, dVdy


def displayf(X, Y, U=None, V=None, background=None, *args):

    if background is not None:
        pylab.pcolor(X, Y, background)
    if U is not None:
        pylab.quiver(X, Y, U, V)


class DataObject:
    pass


class PIV_Postproc(LightPIVResults):
    def __init__(self, path):
        super().__init__(str_path=path)

        warn(
            "The class PIV_Postproc is deprecated (the code is not tested!). "
            "Please use fluidimage.postproc.vector_field.VectorFieldOnGrid instead.",
            DeprecationWarning,
        )

        self.path = os.path.abspath(path)

        self.X, self.Y, self.U, self.V = compute_grid(
            self.ixvecs_final,
            self.iyvecs_final,
            self.deltaxs_final,
            self.deltays_final,
        )

    def displayf(self, U=None, V=None, X=None, Y=None, bg=None, *args):
        if bg is None:
            bg = self.compute_norm(U, V)

        if X is None:
            X = self.X
            Y = self.Y

        if U is None:
            U = self.U
            V = self.V

        displayf(X, Y, U=U, V=V, background=bg, *args)

    def spatial_average(self, U):
        U_mean = np.mean(U, (0, 1))
        return U_mean

    def compute_derivatives(self, edge_order=2):
        dUdx, dUdy, dVdx, dVdy = compute_derivatives(
            self.dx, self.dy, self.U, self.V, edge_order=2
        )
        return dUdx, dUdy, dVdx, dVdy

    def compute_rot(self, edge_order=2):
        if not hasattr(self, "dUdx"):
            self.dUdx, self.dUdy, self.dVdx, self.dVdy = self.compute_derivatives(
                edge_order=edge_order
            )
        rot = compute_rot(self.dUdy, self.dVdx)
        return rot

    def compute_div(self, edge_order=2):
        if ~hasattr(self, "dUdx"):
            self.dUdx, self.dUdy, self.dVdx, self.dVdy = self.compute_derivatives(
                edge_order=edge_order
            )
        div = compute_div(self.dUdx, self.dVdy)
        return div

    def compute_ken(self):
        ken = compute_ken(self.U, self.V)
        return ken

    def compute_norm(self, U, V):
        norm = compute_norm(self.U, self.V)
        return norm

    def compute_spatial_fft(self, parseval=False):

        fftU, kx, ky, psdU = compute_2dspectrum(
            self.X, self.Y, self.U, axes=(0, 1)
        )
        fftV, kx, ky, psdV = compute_2dspectrum(
            self.X, self.Y, self.V, axes=(0, 1)
        )

        if not hasattr(self, "fft"):
            self.fft = DataObject()
        if not hasattr(self.fft, "spatial"):
            spatial = DataObject()
            self.fft.spatial = spatial

        self.fft.spatial.kx = kx
        self.fft.spatial.ky = ky
        self.fft.spatial.fftU = fftU
        self.fft.spatial.fftV = fftV
        self.fft.spatial.psdU = psdU
        self.fft.spatial.psdV = psdV
        # if parseval:
        #     dx = self.X[1][0] - self.X[0][0]
        #     dy = self.Y[0][1] - self.Y[0][0]
        #     Lx = np.max(self.X) - np.min(self.X)
        #     Ly = np.max(self.Y) - np.min(self.Y)
        #     dkx = kx[1] - kx[0]
        #     dky = ky[1] - ky[0]
        #     energphys = np.sum(self.U ** 2 + self.V ** 2) * dx * dy / Lx / Ly
        #     energspectral = (
        #         np.sum(self.fft.spatial.psdU + self.fft.spatial.psdV) * dkx * dky
        #     )
        #     print("%%%% PARSEVAL %%%%")
        #     print("np.sum(U**2+V**2) * dx*dy / Lx/Ly =")
        #     print(energphys)
        #     print("np.sum(psd) * dkx*dky=")
        #     print(energspectral)


class PIV_PostProc_serie(LightPIVResults):
    def __init__(self, path=None):

        warn(
            "The class PIV_Postproc_serie is deprecated (the code is not tested!). "
            "Please use fluidimage.postproc.vector_field.ArrayOfVectorFieldOnGrid instead.",
            DeprecationWarning,
        )

        self.path = path
        path0 = path[0]
        super().__init__(str_path=path0)
        for ti, pathi in enumerate(path[1:]):
            tmp = PIV_Postproc(path=pathi)
            self.deltaxs_final = np.vstack(
                [self.deltaxs_final, tmp.deltaxs_final]
            )
            self.deltays_final = np.vstack(
                [self.deltays_final, tmp.deltays_final]
            )
        self.X, self.Y, self.dx, self.dy, self.U, self.V = self.compute_grid()

    def set_time(self, dt):
        nb_files = len(self.path)
        self.t = np.linspace(0, dt * (nb_files - 1), nb_files)

    def displayf(
        self, U=None, V=None, bg=None, X=None, Y=None, timesleep=0.5, *args
    ):
        if U is None and bg.ndim == 3:
            U = V = [None] * len(bg)
        elif bg is None and U.ndim == 3:
            bg = self.compute_norm(U, V)

        if X is None:
            X = self.X
            Y = self.Y

        if U is None:
            displayf(X, Y, U=U, V=V, background=bg, *args)
        else:
            for i in range(len(U)):
                displayf(X, Y, U=U[i], V=V[i], background=bg[i], *args)
                pylab.show()
                time.sleep(timesleep)

    def time_average(self, U):
        U_mean = np.mean(U, 0)
        return U_mean

    def spatial_average(self, U):
        U_mean = np.mean(U, (1, 2))
        return U_mean

    def compute_grid(self):
        U = [None] * len(self.path)
        V = [None] * len(self.path)
        for ti, pathi in enumerate(self.path):
            X, Y, dx, dy, U[ti], V[ti] = compute_grid(
                self.ixvecs_final,
                self.iyvecs_final,
                self.deltaxs_final[ti],
                self.deltays_final[ti],
            )
        return X, Y, dx, dy, U, V

    def compute_derivatives(self, edge_order=2):
        dUdx = np.zeros(np.shape(self.U))
        dUdy = np.zeros(np.shape(self.U))
        dVdx = np.zeros(np.shape(self.U))
        dVdy = np.zeros(np.shape(self.U))
        for ti, pathi in enumerate(self.path):
            dUdx[ti], dUdy[ti], dVdx[ti], dVdy[ti] = compute_derivatives(
                self.dx, self.dy, self.U[ti], self.V[ti], edge_order=2
            )
        return dUdx, dUdy, dVdx, dVdy

    def compute_rot(self, edge_order=2):
        if not hasattr(self, "dUdx"):
            self.dUdx, self.dUdy, self.dVdx, self.dVdy = self.compute_derivatives(
                edge_order=edge_order
            )
        rot = np.zeros(np.shape(self.U))
        for ti, pathi in enumerate(self.path):
            rot[ti] = compute_rot(self.dUdy[ti], self.dVdx[ti])
        return rot

    def compute_div(self, edge_order=2):
        if ~hasattr(self, "dUdx"):
            self.dUdx, self.dUdy, self.dVdx, self.dVdy = self.compute_derivatives(
                edge_order=edge_order
            )
        div = np.zeros(np.shape(self.U))
        for ti, pathi in enumerate(self.path):
            div[ti] = compute_div(self.dUdx[ti], self.dVdy[ti])
        return div

    def compute_ken(self):
        ken = np.zeros(np.shape(self.U))
        for ti, pathi in enumerate(self.path):
            ken[ti] = compute_ken(self.U[ti], self.V[ti])
        return ken

    def compute_norm(self, U, V):
        norm = np.zeros(np.shape(self.U))
        for ti, pathi in enumerate(self.path):
            norm[ti] = compute_norm(U[ti], V[ti])
        return norm

    def compute_temporal_fft(self):
        if not hasattr(self, "t"):
            print("please define time before perform temporal Fourier transform")
        else:
            fftU, omega, psdU = compute_1dspectrum(self.t, self.U, axis=0)
            fftV, omega, psdV = compute_1dspectrum(self.t, self.V, axis=0)

            if not hasattr(self, "fft"):
                self.fft = DataObject()
            if not hasattr(self.fft, "time"):
                time = DataObject()
                self.fft.time = time

            self.fft.time.omega = omega
            self.fft.time.fftU = fftU
            self.fft.time.fftV = fftV
            self.fft.time.psdU = psdU
            self.fft.time.psdV = psdV
            # if parseval:
            #     # parseval
            #     dt = self.t[1] - self.t[0]
            #     Lt = np.max(self.t) - np.min(self.t)
            #     domega = omega[1] - omega[0]
            #     energphys = np.sum(self.U ** 2 + self.V ** 2) * dt / Lt
            #     energspectral = (
            #         np.sum(self.fft.time.psdU + self.fft.time.psdV) * domega
            #     )
            #     print("%%%% PARSEVAL %%%%")
            #     print("np.sum(U**2+V**2) * dt / Lt =")
            #     print(energphys)
            #     print("np.sum(psd) * domega=")
            #     print(energspectral)

    def compute_spatial_fft(self):

        fftU, kx, ky, psdU = compute_2dspectrum(
            self.X, self.Y, self.U, axes=(1, 2)
        )
        fftV, kx, ky, psdV = compute_2dspectrum(
            self.X, self.Y, self.V, axes=(1, 2)
        )

        if not hasattr(self, "fft"):
            self.fft = DataObject()
        if not hasattr(self.fft, "spatial"):
            spatial = DataObject()
            self.fft.spatial = spatial

        self.fft.spatial.kx = kx
        self.fft.spatial.ky = ky
        self.fft.spatial.fftU = fftU
        self.fft.spatial.fftV = fftV
        self.fft.spatial.psdU = psdU
        self.fft.spatial.psdV = psdV
        # if parseval:
        #     dx = self.X[1][0] - self.X[0][0]
        #     dy = self.Y[0][1] - self.Y[0][0]
        #     Lx = np.max(self.X) - np.min(self.X)
        #     Ly = np.max(self.Y) - np.min(self.Y)
        #     dkx = kx[1] - kx[0]
        #     dky = ky[1] - ky[0]
        #     energphys = np.sum(self.U ** 2 + self.V ** 2) * dx * dy / Lx / Ly
        #     energspectral = (
        #         np.sum(self.fft.spatial.psdU + self.fft.spatial.psdV) * dkx * dky
        #     )
        #     print("%%%% PARSEVAL %%%%")
        #     print("np.sum(U**2+V**2) * dx*dy / Lx/Ly =")
        #     print(energphys)
        #     print("np.sum(psd) * dkx*dky=")
        #     print(energspectral)

    def compute_spatiotemp_fft(self):
        if hasattr(self, "fft.spatial"):
            fftU, omega, psdU = compute_1dspectrum(
                self.t, self.fft.spatial.fftU, axis=0
            )
            fftV, omega, psdV = compute_1dspectrum(
                self.t, self.fft.spatial.fftV, axis=0
            )
            kx = self.fft.spatial.kx
            ky = self.fft.spatial.ky
            Lkx = np.max(kx) - np.min(kx)
            nx = np.shape(self.X)[0]
            Lky = np.max(ky) - np.min(ky)
            ny = np.shape(self.Y)[0]
            psdU *= 1.0 / Lkx / nx / Lky / ny
            psdV *= 1.0 / Lkx / nx / Lky / ny
        elif hasattr(self, "fft.time"):
            fftU, kx, ky, psdU = compute_2dspectrum(
                self.X, self.Y, self.fft.time.fftU, axes=(1, 2)
            )
            fftV, kx, ky, psdV = compute_2dspectrum(
                self.X, self.Y, self.fft.time.fftV, axes=(1, 2)
            )
            omega = self.fft.time.omega
            Lomega = np.max(omega) - np.min(omega)
            n = np.shape(self.t)[0]
            psdU *= 1.0 / Lomega / n
            psdV *= 1.0 / Lomega / n
        else:
            self.compute_temporal_fft()
            fftU, kx, ky, psdU = compute_2dspectrum(
                self.X, self.Y, self.fft.time.fftU, axes=(1, 2)
            )
            fftV, kx, ky, psdV = compute_2dspectrum(
                self.X, self.Y, self.fft.time.fftV, axes=(1, 2)
            )
            omega = self.fft.time.omega
            Lomega = np.max(omega) - np.min(omega)
            n = np.shape(self.t)[0]
            psdU *= 1.0 / Lomega / n
            psdV *= 1.0 / Lomega / n

        if not hasattr(self, "fft"):
            self.fft = DataObject()
        if not hasattr(self.fft, "spatiotemp"):
            spatiotemp = DataObject()
            self.fft.spatiotemp = spatiotemp

        self.fft.spatiotemp.omega = omega
        self.fft.spatiotemp.kx = kx
        self.fft.spatiotemp.ky = ky
        self.fft.spatiotemp.fftU = fftU
        self.fft.spatiotemp.fftV = fftV
        self.fft.spatiotemp.psdU = psdU
        self.fft.spatiotemp.psdV = psdV

        # if parseval:
        #     dx = self.X[1][0] - self.X[0][0]
        #     dy = self.Y[0][1] - self.Y[0][0]
        #     Lx = np.max(self.X) - np.min(self.X)
        #     Ly = np.max(self.Y) - np.min(self.Y)
        #     dkx = kx[1] - kx[0]
        #     dky = ky[1] - ky[0]
        #     dt = self.t[1] - self.t[0]
        #     Lt = np.max(self.t) - np.min(self.t)
        #     domega = omega[1] - omega[0]
        #     energphys = (
        #         np.sum(self.U ** 2 + self.V ** 2) * dx * dy * dt / Lx / Ly / Lt
        #     )
        #     energspectral = (
        #         np.sum(self.fft.spatiotemp.psdU + self.fft.spatiotemp.psdV)
        #         * dkx
        #         * dky
        #         * domega
        #     )
        #     print("%%%% PARSEVAL %%%%")
        #     print("np.sum(U**2+V**2) * dx*dy*dt / Lx/Ly/Lt =")
        #     print(energphys)
        #     print("np.sum(psd) * dkx*dky*domega=")
        #     print(energspectral)



"""Smooth 2d fields (:mod:`fluidimage.calcul.smooth_clean`)
===========================================================


"""
import numpy as np
from scipy.interpolate import interp2d
from scipy.ndimage import convolve

from fluidimage.calcul.interpolate.griddata import griddata

weights = np.ones([3, 3])


def _smooth(a, for_norm):
    norm = convolve(for_norm, weights, mode="nearest")
    ind = np.where(norm == 0)
    norm[ind] = 1
    return convolve(a, weights, mode="nearest") / norm


def smooth_clean(xs, ys, deltaxs, deltays, iyvecs, ixvecs, threshold):

    nx = len(ixvecs)
    ny = len(iyvecs)

    shape = [ny, nx]
    for_norm = np.ones(shape)

    selection = ~(np.isnan(deltaxs) | np.isnan(deltays))
    if selection.any():
        centers = np.vstack([xs[selection], ys[selection]])
        dxs_select = deltaxs[selection]
        dys_select = deltays[selection]
        dxs = griddata(centers, dxs_select, (ixvecs, iyvecs)).reshape([ny, nx])
        dys = griddata(centers, dys_select, (ixvecs, iyvecs)).reshape([ny, nx])

        dxs2 = _smooth(dxs, for_norm)
        dys2 = _smooth(dys, for_norm)

        inds = (abs(dxs2 - dxs) + abs(dys2 - dys) > threshold).nonzero()

        dxs[inds] = 0
        dys[inds] = 0
        for_norm[inds] = 0

        dxs_smooth = _smooth(dxs, for_norm)
        dys_smooth = _smooth(dys, for_norm)

        # come back to the unstructured grid
        fxs = interp2d(ixvecs, iyvecs, dxs_smooth, kind="linear")
        fys = interp2d(ixvecs, iyvecs, dys_smooth, kind="linear")

        out_dxs = np.empty_like(deltaxs)
        out_dys = np.empty_like(deltays)

        for i, (x, y) in enumerate(zip(xs, ys)):
            out_dxs[i] = fxs(x, y)[0]
            out_dys[i] = fys(x, y)[0]

    # from fluiddyn import ipydebug
    # import matplotlib.pylab as plt
    # plt.ion()
    # ipydebug()
    else:
        out_dxs, out_dys = deltaxs, deltays
    return out_dxs, out_dys



from . import BaseWork
from ..data_objects.piv import get_slices_from_strcrop


class BaseWorkWithMask(BaseWork):
    def _xymasked_from_xyoriginalimage(self, xs, ys):
        if self.params.mask.strcrop is not None:
            slices = get_slices_from_strcrop(self.params.mask.strcrop)
            if slices[1].start is not None:
                xs = xs - slices[1].start
            if slices[0].start is not None:
                ys = ys - slices[0].start
        return xs, ys

    def _xyoriginalimage_from_xymasked(self, xs, ys):
        if self.params.mask.strcrop is not None:
            slices = get_slices_from_strcrop(self.params.mask.strcrop)
            if slices[1].start is not None:
                xs = xs + slices[1].start
            if slices[0].start is not None:
                ys = ys + slices[0].start
        return xs, ys

    @classmethod
    def _complete_params_with_default_mask(cls, params):

        params._set_child("mask", attribs={"strcrop": None})

        params.mask._set_doc(
            """
Parameters describing how images are masked.

strcrop : None, str

    Two-dimensional slice (for example '100:600, :'). If None, the whole image
    is used.
"""
        )



"""Topology for image preprocessing (:mod:`fluidimage.topologies.preproc`)
==========================================================================

.. autoclass:: TopologyPreproc
   :members:
   :private-members:

"""
import copy
import json
import os
import sys
from typing import Dict, List, Tuple

from fluiddyn.util.paramcontainer import ParamContainer
from fluidimage import SeriesOfArrays
from fluidimage.data_objects.preproc import ArraySerie as ArraySubset
from fluidimage.data_objects.preproc import get_name_preproc
from fluidimage.topologies import TopologyBase, prepare_path_dir_result
from fluidimage.util import DEBUG, imread, logger
from fluidimage.works.preproc import (
    WorkPreproc,
    _make_doc_with_filtered_params_doc,
)

from . import image2image
from .piv import is_name_in_queue


class TopologyPreproc(TopologyBase):
    """Preprocess series of images.

    The most useful methods for the user (in particular :func:`compute`) are
    defined in the base class :class:`fluidimage.topologies.base.TopologyBase`.

    Parameters
    ----------

    params: None

      A ParamContainer (created with the class method
      :func:`create_default_params`) containing the parameters for the
      computation.

    logging_level: str, {'warning', 'info', 'debug', ...}

      Logging level.

    nb_max_workers: None, int

      Maximum numbers of "workers". If None, a number is computed from the
      number of cores detected. If there are memory errors, you can try to
      decrease the number of workers.

    """

    @classmethod
    def create_default_params(cls, backend="python"):
        """Class method returning the default parameters.

        Typical usage::

          params = TopologyPreproc.create_default_params()
          # modify parameters here
          ...

          topo = TopologyPreproc(params)

        Parameters
        ----------

        backend : {'python', 'opencv'}

            Specifies which backend to use.

        """
        params = WorkPreproc.create_default_params(backend)
        params.preproc.series._set_attribs(
            {
                "strcouple": "i:i+1",
                "ind_start": 0,
                "ind_stop": None,
                "ind_step": 1,
            }
        )

        params.preproc.series._set_doc(
            """
Parameters describing image loading prior to preprocessing.

strcouple : str
    Determines the subset from the whole series of images that should be loaded
    and preprocessed together. Particularly useful when temporal filtering requires
    multiple images.

    For example, for a series of images with just one index,

        >>> strcouple = 'i:i+1'   # load one image at a time
        >>> strcouple = 'i-2:i+3'  # loads 5 images at a time

    Similarly for two indices,

        >>> strcouple = 'i:i+1,0'   # load one image at a time, with second index fixed
        >>> strcouple = 'i-2:i+3,0'  # loads 5 images at a time, with second index fixed

    .. todo::

        Rename this parameter to strsubset / strslice

ind_start : int
    Start index for the whole series of images being loaded.
    For more details: see `class SeriesOfArrays`.

ind_stop : int
    Stop index for the whole series of images being loaded.
    For more details: see `class SeriesOfArrays`.

ind_step : int
    Step index for the whole series of images being loaded.
    For more details: see `class SeriesOfArrays`.

"""
        )

        params.preproc._set_child(
            "saving",
            attribs={
                "path": None,
                "strcouple": None,
                "how": "ask",
                "format": "img",
                "postfix": "pre",
            },
        )

        params.preproc.saving._set_doc(
            """
Parameters describing image saving after preprocessing.

path : str or None
    Path to which preprocessed images are saved.

str_subset : str or None
    NotImplemented! Determines the sub-subset of images must be saved from subset
    of images that were loaded and preprocessed. When set as None, saves the
    middle image from every subset.

    .. todo::

        Implement the option params.saving.str_subset...

how : str {'ask', 'new_dir', 'complete', 'recompute'}
    How preprocessed images must be saved if it already exists or not.

format : str {'img', 'hdf5'}
    Format in which preprocessed image data must be saved.

postfix : str
    A suffix added to the new directory where preprocessed images are saved.

"""
        )

        params._set_internal_attr(
            "_value_text",
            json.dumps(
                {
                    "program": "fluidimage",
                    "module": "fluidimage.topologies.preproc",
                    "class": "TopologyPreproc",
                }
            ),
        )

        params._set_child("im2im")
        image2image.complete_im2im_params_with_default(params.im2im)

        return params

    def __init__(
        self, params: ParamContainer, logging_level="info", nb_max_workers=None
    ):
        self.params = params.preproc

        self.preproc_work = WorkPreproc(params)
        self.results = []
        self.display = self.preproc_work.display

        serie_arrays = self.preproc_work.serie_arrays
        self.series = SeriesOfArrays(
            serie_arrays,
            params.preproc.series.strcouple,
            ind_start=params.preproc.series.ind_start,
            ind_stop=params.preproc.series.ind_stop,
            ind_step=params.preproc.series.ind_step,
        )

        subset = self.series.get_serie_from_index(0)
        self.nb_items_per_serie = subset.get_nb_arrays()

        if os.path.isdir(params.preproc.series.path):
            path_dir = params.preproc.series.path
        else:
            path_dir = os.path.dirname(params.preproc.series.path)
        self.path_dir_input = path_dir

        path_dir_result, self.how_saving = prepare_path_dir_result(
            path_dir,
            params.preproc.saving.path,
            params.preproc.saving.postfix,
            params.preproc.saving.how,
        )

        super().__init__(
            path_dir_result=path_dir_result,
            logging_level=logging_level,
            nb_max_workers=nb_max_workers,
        )

        self.params.saving.path = self.path_dir_result

        # Define waiting queues
        queue_subsets_of_names = self.add_queue("subsets of filenames")
        queue_paths = self.add_queue("image paths")
        queue_arrays = queue_arrays1 = self.add_queue("arrays")
        queue_subsets_of_arrays = self.add_queue("subsets of arrays")
        queue_preproc_objects = self.add_queue("preproc results")

        if params.im2im.im2im is not None:
            queue_arrays1 = self.add_queue("arrays1")

        # Define works
        self.add_work(
            "fill (subsets_of_names, paths)",
            func_or_cls=self.fill_subsets_of_names_and_paths,
            output_queue=(queue_subsets_of_names, queue_paths),
            kind=("global", "one shot"),
        )

        self.add_work(
            "imread",
            func_or_cls=imread,
            input_queue=queue_paths,
            output_queue=queue_arrays,
            kind="io",
        )

        if params.im2im.im2im is not None:
            im2im_func = image2image.TopologyImage2Image.init_im2im(
                self, params.im2im
            )

            self.add_work(
                "image2image",
                func_or_cls=im2im_func,
                input_queue=queue_arrays,
                output_queue=queue_arrays1,
            )

        self.add_work(
            "make subsets of arrays",
            func_or_cls=self.make_subsets,
            input_queue=(queue_subsets_of_names, queue_arrays1),
            output_queue=queue_subsets_of_arrays,
            kind="global",
        )

        self.add_work(
            "preproc a subset of arrays",
            func_or_cls=self.preproc_work.calcul,
            params_cls=params,
            input_queue=queue_subsets_of_arrays,
            output_queue=queue_preproc_objects,
        )

        self.add_work(
            "save images",
            func_or_cls=self.save_preproc_object,
            input_queue=queue_preproc_objects,
            kind="io",
        )

    def save_preproc_object(self, obj: ArraySubset):
        """Save a preprocessing object"""
        ret = obj.save(path=self.path_dir_result)
        self.results.append(ret)

    def init_series(self) -> List[str]:
        """Initializes the SeriesOfArrays object `self.series` based on input
        parameters."""
        series = self.series
        if not series:
            logger.warning("encountered empty series. No images to preprocess.")
            return

        if self.how_saving == "complete":
            index_subsets = []
            for ind_subset, subset in self.series.items():
                names_serie = subset.get_name_arrays()
                name_preproc = get_name_preproc(
                    subset,
                    names_serie,
                    ind_subset,
                    series.nb_series,
                    self.params.saving.format,
                )
                if not (self.path_dir_result / name_preproc).exists():
                    index_subsets.append(ind_subset)
            series.set_index_series(index_subsets)
            if logger.isEnabledFor(DEBUG):
                logger.debug(
                    repr([subset.get_name_arrays() for subset in series])
                )

        nb_subsets = len(series)
        if nb_subsets == 0:
            logger.warning('topology in mode "complete" and work already done.')
            return
        elif nb_subsets == 1:
            plurial = ""
        else:
            plurial = "s"

        logger.info(f"Add {nb_subsets} image serie{plurial} to compute.")

    def fill_subsets_of_names_and_paths(
        self, input_queue: None, output_queues: Tuple[Dict]
    ) -> None:
        """Fill the two first queues"""
        assert input_queue is None
        queue_subsets_of_names, queue_paths = output_queues

        self.init_series()

        for ind_subset, subset in self.series.items():
            queue_subsets_of_names[ind_subset] = subset.get_name_arrays()
            for name, path in subset.get_name_path_arrays():
                queue_paths[name] = path

    def make_subsets(self, input_queues: Tuple[Dict], output_queue: Dict) -> bool:
        """Create the subsets of images"""
        queue_subsets_of_names, queue_arrays = input_queues
        # for each name subset
        for key, names in list(queue_subsets_of_names.items()):
            # if correspondant arrays have been loaded from images,
            # make an array subset
            if all([name in queue_arrays for name in names]):
                arrays = (queue_arrays[name] for name in names)
                serie = copy.copy(self.series.get_serie_from_index(key))

                array_subset = ArraySubset(
                    names=names, arrays=arrays, serie=serie
                )
                output_queue[key] = array_subset
                del queue_subsets_of_names[key]
                # remove the image_array if it not will be used anymore

                key_arrays = list(queue_arrays.keys())
                for key_array in key_arrays:
                    if not is_name_in_queue(key_array, queue_subsets_of_names):
                        del queue_arrays[key_array]


Topology = TopologyPreproc

if "sphinx" in sys.modules:
    __doc__ += _make_doc_with_filtered_params_doc(Topology)



"""Graphical user interface
===========================

Two simple graphical user interfaces (GUI) can be launched with the commands
`fluidimviewer` and `fluidimlauncher`.

This sub-package contains the code for the fluidimage GUI:

.. autosummary::
   :toctree:

   imviewer
   launcher
   pg_main

"""



"""Thin plate spline
====================

Translated and adapted from UVmat code (Joel Sommeria, LEGI, CNRS).

This interpolation/smoothing (Duchon, 1976; NguyenDuc and Sommeria,
1988) minimises a linear combination of the squared curvature and
squared difference from the initial data.

We first need to compute tps coefficients ``U_tps`` (function
``compute_tps_coeff``). Interpolated data can then be obtained as the
matrix product ``dot(U_tps, EM)`` where the matrix ``EM`` is obtained
by the function ``compute_tps_matrix``.  The spatial derivatives are
obtained as ``dot(U_tps, EMDX)`` and ``dot(U_tps, EMDY)``, where
``EMDX`` and ``EMDY`` are obtained from the function
``compute_tps_matrix_dxy``. A helper class is also provided.

.. autofunction:: compute_tps_coeff

.. autoclass:: ThinPlateSpline
   :members:

.. autofunction:: compute_tps_matrix_numpy

.. autofunction:: compute_tps_matrices_dxy


"""

import numpy as np

from transonic import Transonic, boost

ts = Transonic()

A = "float64[][]"


@boost
def compute_tps_matrix_pythran(new_pos: A, centers: A):
    """calculate the thin plate spline (tps) interpolation at a set of points

    Parameters
    ----------

    new_pos: np.array
        ``[nb_dim, M]`` array representing the postions of the M
        'observation' sites, with nb_dim the space dimension.

    centers: np.array
        ``[nb_dim, N]`` array representing the postions of the N centers,
        sources of the tps.

    Returns
    -------

    EM : np.array
        ``[(N+nb_dim), M]`` matrix representing the contributions at the M sites.

        From unit sources located at each of the N centers, +
        (nb_dim+1) columns representing the contribution of the linear
        gradient part.

    Notes
    -----

    >>> U_interp = np.dot(U_tps, EM)

    """

    d, nb_new_pos = new_pos.shape
    d2, nb_centers = centers.shape
    assert d == d2

    EM = np.zeros((nb_centers, nb_new_pos))

    # # pythran 0.9.2 does not know np.meshgrid
    # for ind_d in range(s):
    #     Dsites, Centers = np.meshgrid(dsites[ind_d], centers[ind_d])
    #     EM += (Dsites - Centers)**2

    for ind_d in range(d):
        for ic, center in enumerate(centers[ind_d]):
            for inp, npos in enumerate(new_pos[ind_d]):
                EM[ic, inp] += (npos - center) ** 2

    # Pythran does not like that!
    # nb_p = np.where(EM != 0)
    # EM[nb_p] = EM[nb_p] * np.log(EM[nb_p]) / 2

    for ic in range(nb_centers):
        for inp in range(nb_new_pos):
            tmp = EM[ic, inp]
            if tmp != 0:
                EM[ic, inp] = tmp * np.log(tmp) / 2

    # # pythran 0.9.2 does not know np.vstack
    # EM_ret = np.vstack([EM, np.ones(nb_new_pos), new_pos])

    EM_ret = np.empty((nb_centers + 1 + d, nb_new_pos))
    EM_ret[:nb_centers, :] = EM
    EM_ret[nb_centers, :] = np.ones(nb_new_pos)
    EM_ret[nb_centers + 1 : nb_centers + 1 + d, :] = new_pos

    return EM_ret


def compute_tps_matrix_numpy(new_pos: A, centers: A):
    """calculate the thin plate spline (tps) interpolation at a set of points

    Parameters
    ----------

    new_pos: np.array
        ``[nb_dim, M]`` array representing the postions of the M
        'observation' sites, with nb_dim the space dimension.

    centers: np.array
        ``[nb_dim, N]`` array representing the postions of the N centers,
        sources of the tps.

    Returns
    -------

    EM : np.array

        ``[(N+nb_dim), M]`` matrix representing the contributions at the M
        sites.

        From unit sources located at each of the N centers, +
        (nb_dim+1) columns representing the contribution of the linear
        gradient part.

    Notes
    -----

    >>> U_interp = np.dot(U_tps, EM)

    """

    d, nb_new_pos = new_pos.shape
    d2, nb_centers = centers.shape
    assert d == d2

    EM = np.zeros((nb_centers, nb_new_pos))

    for ind_d in range(d):
        Dsites, Centers = np.meshgrid(new_pos[ind_d], centers[ind_d])
        EM += (Dsites - Centers) ** 2

    nb_p = np.where(EM != 0)
    EM[nb_p] = EM[nb_p] * np.log(EM[nb_p]) / 2
    EM_ret = np.vstack([EM, np.ones(nb_new_pos), new_pos])
    return EM_ret


if ts.is_compiled:

    def compute_tps_matrix(newcenters, centers):
        return compute_tps_matrix_pythran(
            newcenters.astype(np.float64), centers.astype(np.float64)
        )

else:
    print("Warning: function compute_tps_matrix_numpy not pythranized.")
    compute_tps_matrix = compute_tps_matrix_numpy


def compute_tps_coeff(centers, U, smoothing_coef):
    """Calculate the thin plate spline (tps) coefficients

    Parameters
    ----------

    centers : np.array
        ``[nb_dim,  N]`` array representing the positions of the N centers,
        sources of the TPS (nb_dim = space dimension).

    U : np.array
        ``[N]`` array representing the values of the considered
        scalar measured at the centres ``centers``.

    smoothing_coef : float
        Smoothing parameter. The result is smoother for larger smoothing_coef.

    Returns
    -------

    U_smooth : np.array
         Values of the quantity U at the N centres after smoothing.

    U_tps : np.array
         TPS weights of the centres and columns of the linear.

    """
    nb_dim, N = centers.shape
    U = np.hstack([U, np.zeros(nb_dim + 1)])
    U = U.reshape([U.size, 1])
    try:
        EM = compute_tps_matrix(centers, centers).T
    except TypeError as e:
        print(centers.dtype, centers.shape)
        raise e

    smoothing_mat = smoothing_coef * np.eye(N, N)
    smoothing_mat = np.hstack([smoothing_mat, np.zeros([N, nb_dim + 1])])
    PM = np.hstack([np.ones([N, 1]), centers.T])
    IM = np.vstack(
        [
            EM + smoothing_mat,
            np.hstack([PM.T, np.zeros([nb_dim + 1, nb_dim + 1])]),
        ]
    )

    # print('det(IM)', np.linalg.det(IM))
    # print('cond(IM)', np.linalg.cond(IM))

    # U_tps, r, r2, r3 = np.linalg.lstsq(IM, U)
    U_tps = np.linalg.solve(IM, U)

    U_smooth = np.dot(EM, U_tps)
    return U_smooth.ravel(), U_tps.ravel()


def compute_tps_matrices_dxy(dsites, centers):
    """Calculate the derivatives of thin plate spline (tps) interpolation
    at a set of points (limited to the 2D case)

    Parameters
    ----------

    dsites : np.array
        ``[nb_dim,  M]`` array of interpolation site coordinates
        (nb_dim = space dimension = 2, here).

    centers : np.array
        ``[nb_dim,  N]`` array of centre coordinates (initial data).

    Returns
    -------

    DMX : np.array
        ``[(N+3),  M]`` array representing the contributions to the X
        derivatives at the M sites from unit sources located at each
        of the N centers, + 3 columns representing the contribution of
        the linear gradient part.

    DMY : np.array
        idem for Y derivatives.

    """
    s, M = dsites.shape
    s2, N = centers.shape
    assert s == s2
    Dsites, Centers = np.meshgrid(dsites[1], centers[1])
    DX = Dsites - Centers
    Dsites, Centers = np.meshgrid(dsites[0], centers[0])
    DY = Dsites - Centers
    DM = DX * DX + DY * DY
    DM[DM != 0] = np.log(DM[DM != 0]) + 1
    DMX = np.vstack([DX * DM, np.zeros(M), np.ones(M), np.zeros(M)])
    DMY = np.vstack([DY * DM, np.zeros(M), np.zeros(M), np.ones(M)])
    return DMX, DMY


class ThinPlateSpline:
    """Helper class for thin plate interpolation."""

    _compute_tps_matrix = compute_tps_matrix

    def __init__(self, new_positions, centers):
        self.EM = type(self)._compute_tps_matrix(new_positions, centers)
        self.DMX, self.DMY = compute_tps_matrices_dxy(new_positions, centers)

    def compute_field(self, U_tps):
        """Compute the interpolated field."""
        return np.dot(U_tps, self.EM)

    def compute_gradient(self, U_tps):
        """Compute the gradient (dx_U, dy_U)"""
        return np.dot(U_tps, self.DMX), np.dot(U_tps, self.DMY)


class ThinPlateSplineNumpy(ThinPlateSpline):
    pass
    # _compute_tps_matrix = compute_tps_matrix_numpy



"""Tomography data objects (:mod:`fluidimage.data_objects.tomo`)
================================================================

.. autoclass:: ArrayTomoBase
   :members:
   :private-members:

.. autoclass:: ArrayTomo
   :members:
   :private-members:

.. autoclass:: ArrayTomoCV
   :members:
   :private-members:

"""


import os
import shutil

import dask.array as da
import h5py
import matplotlib.pyplot as plt
import numpy as np

try:
    import ipyvolume.pylab as p3
except ImportError:
    print(
        "Install ipyvolume to visualize ArrayTomo in 3D in a Jupyter widget." ""
    )


class ArrayTomoBase:
    """Data structure to hold the tomographic data for a single instant."""

    _keys_to_save = ["xs", "ys", "zs", "I"]
    _attrs_to_save = ["nx", "ny", "nz"]

    def __init__(
        self,
        xlims=(0, 10),
        ylims=(0, 10),
        zlims=(0, 5),
        nb_voxels=(10, 10, 5),
        image_path=None,
        params=None,
        h5file_path=None,
        dtype=np.float64,
    ):
        """Initialize

        Parameters
        ----------
        xlims, ylims, zlims : tuple of int
            Limits of the 3D physical volume geometry

        nb_voxels : tuple of int or array-like
            Shape of the voxel array in the three axes, ie. (nx, ny, nz)

        image_path : str
            Reference image filename

        dtype : numpy floating point datatype
            Control the precision of the arrays

        """

        self.xmin, self.xmax = xlims
        self.ymin, self.ymax = ylims
        self.ymin, self.zmax = zlims
        self.nb_voxels = nb_voxels
        self.dtype = dtype
        if h5file_path is not None:
            self.load(h5file_path)
        else:
            self.nx, self.ny, self.nz = nb_voxels

            def grid1d(lims, nb):
                start, stop = lims
                return np.linspace(start, stop, nb, endpoint=True, dtype=dtype)

            self.xs, self.ys, self.zs = map(
                grid1d, (xlims, ylims, zlims), nb_voxels
            )

        if image_path is not None:
            self.init_paths(image_path)
        self.params = params

    def init_paths(self, image_path, output_dir=None):
        self.image_path = image_path
        if output_dir is None:
            output_dir = os.path.abspath(
                os.path.join(os.path.dirname(image_path), "..", "tomo")
            )
        self.h5file_path = os.path.join(output_dir, self._get_name())

    def describe(self, vmin=None, vmax=None):
        """Describe the voxel intensity array."""
        I = self.I
        if vmin is not None:
            x, y = np.where(I < vmin)
            print(f"No. of points below {vmin}={x.size}")
        if vmax is not None:
            x, y = np.where(I > vmax)
            print(f"No. of points above {vmax}={x.size}")
        string = (
            f" min={I.min()}\n max={I.max()}\n mean={I.mean()}\n median={np.median(I)}\n "
            f" std={I.std()}\n shape={I.shape}"
        )
        string = [s.split("=") for s in string.splitlines()]
        final_string = []
        for l, r in string:
            final_string.append(" = ".join((l.rjust(10), r)))
        print("\n".join(final_string))

    def clear(self):
        """Reset intensities of all voxels as unity."""
        if isinstance(self.I, np.ndarray):
            self.I[:] = 1
        else:
            print(f"Warning: Not clearing I of type {type(self.I)}")
        if os.path.exists(self.h5file_path):
            print(f"rm {self.h5file_path}")
            os.remove(self.h5file_path)

    def load(self, h5file_path=None):
        if h5file_path is None:
            h5file_path = self.h5file_path
        else:
            self.h5file_path = h5file_path

        with h5py.File(h5file_path, "r") as file:
            grp = file["tomo"]
            for k in self._keys_to_save:
                setattr(self, k, grp[k][...])
            for attr in self._attrs_to_save:
                setattr(self, attr, grp.attrs[attr])

    def load_dataset(self, h5file_path=None, tag="tomo", key="I", copy=False):
        if h5file_path is None:
            if copy:
                h5file_path, ext = os.path.splitext(self.h5file_path)
                h5file_path += "_tmp" + ext
                print(f"Copying {self.h5file_path} to {h5file_path}...", end="")
                shutil.copyfile(self.h5file_path, h5file_path)
                print("Done.")
            else:
                h5file_path = self.h5file_path

        file = h5py.File(h5file_path, "r")
        try:
            grp = file[tag]
            dset = grp[key]
        except KeyError:
            print(
                h5file_path,
                "contains\n Groups:",
                list(file),
                "and",
                tag,
                "contains\n Datasets:",
                list(grp),
            )
            file.close()
            raise
        return file, dset, h5file_path

    def _get_name(self):
        return os.path.splitext(os.path.basename(self.image_path))[0] + ".h5"

    def save(self, path=None, sparse=False):
        if path is None:
            path = self.h5file_path
            os.makedirs(os.path.dirname(path), exist_ok=True)

        if sparse:
            raise NotImplementedError("Save the intensity values as sparse")

        with h5py.File(path, "w") as file:
            self._save_in_hdf5_object(file)

    def _save_in_hdf5_object(self, file, tag="tomo"):
        if "class_name" not in file.attrs.keys():
            file.attrs["class_name"] = self.__class__.__name__
            file.attrs["module_name"] = self.__module__
        if "params" not in file.keys() and self.params is not None:
            self.params._save_as_hdf5(hdf5_parent=file)

        if tag in file:
            grp = file[tag]
        else:
            grp = file.create_group(tag)

        grp.attrs["class_name"] = self.__class__.__name__
        grp.attrs["module_name"] = self.__module__
        for attr in self._attrs_to_save:
            grp.attrs[attr] = getattr(self, attr)

        for k in self._keys_to_save:
            data = getattr(self, k)
            print(f"Saving {type(data)} {k}...")
            if isinstance(data, da.core.Array):
                dataset = grp.require_dataset(
                    f"/{tag}/{k}", shape=data.shape, dtype=data.dtype
                )
                da.store(data, dataset)
            else:
                grp.create_dataset(k, data=data)

    def plot3d(self, threshold=0.5):
        """Display the reconstructed intensities as an Jupyter widget using
        ipyvolume.

        """
        if isinstance(self.I, da.core.Array):
            raise ValueError("Cannot display dask arrays.")
        sel = self.I > threshold
        print(f"Displaying {np.where(sel)[0].size} points...")
        fig = p3.figure()
        cmap = plt.cm.gray
        color = cmap(self.I[sel] / self.I.max())
        try:
            s = p3.scatter(
                self.X[sel],
                self.Y[sel],
                self.Z[sel],
                color=color,
                marker="sphere",
            )
        except AttributeError:
            grid_sel = self.grid[sel]
            s = p3.scatter(
                grid_sel[:, 0],
                grid_sel[:, 1],
                grid_sel[:, 2],
                color=color,
                marker="sphere",
            )
        s.size = self.I[sel] / self.I.max() * 5
        p3.view(90, 0)

        return fig, s

    def plot_slices(self, start=0, stop=None):
        I = self.I
        if I.ndim == 1:
            nx, ny, nz = map(len, (self.xs, self.ys, self.zs))
            I = I.reshape((nx, ny, nz))

        if stop is None:
            stop = I.shape[2]

        plt.figure()
        im = plt.imshow(I[:, :, start])
        plt.colorbar(im)
        for i2 in range(start + 1, stop):
            im.set_data(I[:, :, i2])
            plt.draw()
            plt.pause(0.5)


class ArrayTomo(ArrayTomoBase):
    """A typical 3D meshgrid for tomographic reconstruction."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.X, self.Y, self.Z = np.meshgrid(self.xs, self.ys, self.zs)

        # Intensities
        if "h5file_path" not in kwargs:
            self.I = np.ones_like(self.X)


class ArrayTomoCV(ArrayTomoBase):
    """A flattened 3D array for tomographic reconstruction. This format is
    suitable for OpenCV based algorithms.

    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        grid = (
            np.vstack(np.meshgrid(self.zs, self.ys, self.xs, indexing="ij"))
            .reshape(3, -1)
            .T
        )

        # Swap columns to make it back to x, y ,z
        self.grid = np.empty_like(grid)
        self.grid[:, 0] = grid[:, 2]
        self.grid[:, 1] = grid[:, 1]
        self.grid[:, 2] = grid[:, 0]

        if "h5file_path" not in kwargs:
            self.I = np.ones(len(self.grid))



"""Topologies representing asynchronous computations
====================================================

A topology represents an asynchronous computation as a graph of **waiting
queues** and **works**. Works take (an) item(s) in (an) input queue(s), process
it/them and put the result in (an) output queue(s).

All topologies inherit from a base class
:class:`fluidimage.topologies.base.TopologyBase`, which has methods to define
the topology (``add_queue`` and ``add_work``), to represent the computational
graph (``make_code_graphviz``) and finally to execute it (``compute``).

Users are particularly concerned with the following already defined topologies:

.. autosummary::
   :toctree:

   piv
   bos
   preproc
   image2image
   surface_tracking
   optical_flow

These others modules defined classes and functions useful for developers.

.. autosummary::
   :toctree:

   base
   log
   launcher

"""

import os
import sys
from pathlib import Path

from fluiddyn.io.query import query

from .base import TopologyBase
from .log import LogTopology

__all__ = ["LogTopology", "TopologyBase"]

how_values = ("ask", "new_dir", "complete", "recompute")


def prepare_path_dir_result(
    path_dir_input, path_saving, postfix_saving, how_saving
):
    """Makes new directory for results, if required, and returns its path."""

    if how_saving not in how_values:
        raise ValueError(
            f"how_saving (here equal to '{how_saving}') "
            f"should be in {how_values}"
        )

    path_dir_input = str(path_dir_input)

    if path_saving is not None:
        path_dir_result = path_saving
    else:
        path_dir_result = path_dir_input + "." + postfix_saving

    how = how_saving
    if os.path.exists(path_dir_result):
        if how == "ask":
            answer = query(
                f"The directory {path_dir_result} "
                + "already exists. What do you want to do?\n"
                "New dir, Complete, Recompute or Stop?\n"
            )

            while answer.lower() not in ["n", "c", "r", "s"]:
                answer = query(
                    "The answer should be in ['n', 'c', 'r', 's']\n"
                    "Please type your answer again...\n"
                )

            if answer == "s":
                print("Stopped by the user.")
                sys.exit()

            elif answer == "n":
                how = "new_dir"
            elif answer == "c":
                how = "complete"
            elif answer == "r":
                how = "recompute"

        if how == "new_dir":
            i = 0
            while os.path.exists(path_dir_result + str(i)):
                i += 1
            path_dir_result += str(i)

    path_dir_result = Path(path_dir_result)
    path_dir_result.mkdir(exist_ok=True)
    return path_dir_result, how



"""Griddata wrapper function
============================


"""
import numpy as np
from matplotlib import mlab
from scipy import interpolate


def griddata(centers, values, new_positions, using="scipy"):
    xnew = new_positions[0]
    ynew = new_positions[1]
    grid_x, grid_y = np.meshgrid(xnew, ynew)

    if using == "scipy":
        # 'linear' or 'cubic'
        values_new = interpolate.griddata(
            centers.T, values, (grid_x, grid_y), "cubic"
        )
    elif using == "matplotlib":
        x = centers[0]
        y = centers[1]
        values_new = mlab.griddata(x, y, values, xnew, ynew, "linear")
        values_new[values_new.mask] = np.nan
    else:
        raise ValueError

    inds = np.where(np.isnan(values_new))
    values_nearest = interpolate.griddata(
        centers.T, values, (grid_y, grid_x), "nearest"
    )
    values_new[inds] = values_nearest[inds]

    return values_new.flatten()



import os
import re
from multiprocessing import cpu_count
from warnings import warn

from fluidimage.config import get_config

config = get_config()

if "OMP_NUM_THREADS" not in os.environ:
    warn("OMP_NUM_THREADS not set")
else:
    OMP_NUM_THREADS = int(os.environ["OMP_NUM_THREADS"])
    if OMP_NUM_THREADS > 1:
        warn("OMP_NUM_THREADS is greater than 1!")

nb_cores = cpu_count()

allow_hyperthreading = False

if config is not None:
    try:
        allow_hyperthreading = eval(config["topology"]["allow_hyperthreading"])
    except KeyError:
        pass

try:  # should work on UNIX
    # found in http://stackoverflow.com/questions/1006289/how-to-find-out-the-number-of-cpus-using-python # noqa
    with open("/proc/self/status") as file:
        status = file.read()
    m = re.search(r"(?m)^Cpus_allowed:\s*(.*)$", status)
    if m:
        nb_cpus_allowed = bin(int(m.group(1).replace(",", ""), 16)).count("1")

    if nb_cpus_allowed > 0:
        nb_cores = nb_cpus_allowed

    with open("/proc/cpuinfo") as file:
        cpuinfo = file.read()

    nb_proc_tot = 0
    siblings = None
    for line in cpuinfo.split("\n"):
        if line.startswith("processor"):
            nb_proc_tot += 1
        if line.startswith("siblings") and siblings is None:
            siblings = int(line.split()[-1])

    if nb_cores == siblings * 2:
        if not allow_hyperthreading:
            print("We do not use hyperthreading.")
            nb_cores //= 2

except IOError:
    pass



"""Utilities for numerical computation
======================================

This subpackage contains specialized functions and classes for efficient
calculations.

.. autosummary::
   :toctree:

   fft
   correl
   subpix
   interpolate
   smooth_clean

"""



import numpy as np
import pylab

from fluidimage.calcul.correl import correlation_classes
from fluidimage.synthetic import make_synthetic_images

classes = {k.replace(".", "_"): v for k, v in correlation_classes.items()}

try:
    from reikna.cluda import any_api

    api = any_api()
except Exception:
    classes.pop("cufft")

try:
    import pycuda
except ImportError:
    classes.pop("pycuda")

try:
    import skcuda
except ImportError:
    classes.pop("skcufft")

try:
    import theano
except ImportError:
    classes.pop("theano")


def setUpImages(nx, ny, dx, dy, part_size):

    displacements = np.array([dx, dy])

    nb_particles = (nx // 4) ** 2

    im0, im1 = make_synthetic_images(
        displacements,
        nb_particles,
        shape_im0=(ny, nx),
        epsilon=0.0,
        part_size=part_size,
    )

    return displacements, im0, im1


def compute_displacement(cls, im0, im1, nsubpix, method_subpix):
    correl = cls(
        im0.shape, im1.shape, method_subpix=method_subpix, nsubpix=nsubpix
    )

    c, norm = correl(im0, im1)
    dx, dy, correl_max = correl.compute_displacement_from_correl(
        c, coef_norm=norm
    )
    displacement_computed = np.array([dx, dy])

    return displacement_computed


def error_vs_nsubpix(nx, ny, dx, dy, method_subpix, n_subpix, part_size):
    displacements, im0, im1 = setUpImages(nx, ny, dx, dy, part_size)
    # pylab.imshow(im0)
    # pylab.show()

    errdx = np.zeros([np.shape(classes.items())[0], n_subpix.size])
    errdy = np.zeros([np.shape(classes.items())[0], n_subpix.size])

    temp = np.reshape(classes.items(), [6, 2])
    leg = temp.T[0]

    indn = 0
    for nsubpix in n_subpix:
        indk = 0
        for k, cls in classes.items():
            try:
                displacement_computed = compute_displacement(
                    cls, im0, im1, nsubpix, method_subpix
                )
            except:
                displacement_computed = np.array([np.nan, np.nan])
            errdx[indk][indn] = np.abs(displacement_computed - displacements)[0]
            errdy[indk][indn] = np.abs(displacement_computed - displacements)[1]

            indk += 1
        indn += 1

    pylab.figure
    for i, legi in enumerate(leg):
        pylab.plot(n_subpix, np.sqrt(errdx[i] ** 2 + errdy[i] ** 2), "o")
    pylab.legend(leg)
    pylab.xlabel("nsubpix")
    pylab.ylabel("error in pix")
    title = method_subpix + f" _part_size={part_size}" + f" _nx_ny = {nx}_{ny}"
    pylab.title(title)
    pylab.xlim([0, np.max(n_subpix) + 1])
    pylab.show()
    pylab.savefig("./plot_evaluate_subpix/" + title + ".png")

    return errdx, errdy


if __name__ == "__main__":
    part_size = 4.0
    nx = ny = 64
    dx = 3.3
    dy = 3.7
    # method_subpix = '2d_gaussian'
    method_subpix = "centroid"
    n_subpix = np.arange(1, 8)
    errdx, errdy = error_vs_nsubpix(
        nx, ny, dx, dy, method_subpix, n_subpix, part_size
    )



"""FFT classes
==============

Warning: it is more efficient to use not normalized FFT, so we'll do
that.

.. autoclass:: CUFFT2DReal2Complex
   :members:
   :private-members:

.. autoclass:: CUFFT2DReal2ComplexFloat64
   :members:
   :private-members:

.. autoclass:: SKCUFFT2DReal2Complex
   :members:
   :private-members:

.. autoclass:: FFTW2DReal2Complex
   :members:
   :private-members:

.. autoclass:: FFTW2DReal2ComplexFloat64
   :members:
   :private-members:

"""

import numpy as np

try:
    import pyfftw
except ImportError:
    pass

try:
    from reikna.cluda import any_api, cuda_api, ocl_api
    from reikna.fft import FFT
    from reikna.transformations import mul_param
except ImportError:
    pass

try:
    import pycuda._driver
except ImportError:
    pass
else:
    try:
        import pycuda.autoinit
        import pycuda.gpuarray as gpuarray
        import skcuda.fft as skfft
    except (ImportError, pycuda._driver.RuntimeError):
        pass

# if 'OMP_NUM_THREADS' in os.environ:
#     nthreads = int(os.environ['OMP_NUM_THREADS'])
# else:
#     pass

# It seems that it is better to used nthreads = 1 for the fft with
# small size used for PIV
nthreads = 1


class CUFFT2DReal2Complex:
    """A class to use cufft with float32."""

    type_real = "float32"
    type_complex = "complex64"

    def __init__(self, nx, ny):

        shapeX = [ny, nx]
        shapeK = [ny, nx]

        self.shapeX = shapeX
        self.arrayK = np.empty(shapeK, dtype=self.type_complex)

        # Pick the first available GPGPU API and make a Thread on it.
        api = any_api()
        # api = cuda_api()
        # api = ocl_api()
        dev = api.get_platforms()[0].get_devices()
        self.thr = api.Thread.create(dev)
        fft = FFT(self.arrayK, axes=(0, 1))
        scale = mul_param(self.arrayK, np.float)
        fft.parameter.input.connect(
            scale, scale.output, input_prime=scale.input, param=scale.param
        )
        self.fftplan = fft.compile(self.thr, fast_math=True)

        self.coef_norm = nx * ny

    def fft(self, ff):
        arr_dev = self.thr.to_device(ff.astype(self.type_complex))
        self.fftplan(arr_dev, arr_dev, 1.0 / self.coef_norm)
        return arr_dev.get()

    def ifft(self, ff_fft):
        arr_dev = self.thr.to_device(ff_fft)
        self.fftplan(arr_dev, arr_dev, self.coef_norm, inverse=True)
        return arr_dev.get()

    def compute_energy_from_Fourier(self, ff_fft):
        return np.sum(abs(ff_fft) ** 2) / 2

    def compute_energy_from_spatial(self, ff):
        return np.mean(abs(ff) ** 2) / 2

    def project_fft_on_realX(self, ff_fft):
        return self.fft(self.ifft(ff_fft))


class CUFFT2DReal2ComplexFloat64(CUFFT2DReal2Complex):
    """A class to use cufft with float64."""

    type_real = "float64"
    type_complex = "complex128"


class SKCUFFT2DReal2Complex:
    """A class to use skcuda-cufft with float32."""

    type_real = "float32"
    type_complex = "complex64"

    def __init__(self, nx, ny):

        shapeX = [ny, nx]
        shapeK = [ny, nx // 2 + 1]

        self.shapeX = shapeX
        self.shapeK = shapeK

        self.fftplan = skfft.Plan(self.shapeX, np.float32, np.complex64)
        self.ifftplan = skfft.Plan(self.shapeX, np.complex64, np.float32)

        self.coef_norm = nx * ny

    def fft(self, ff):
        x_gpu = gpuarray.to_gpu(ff)
        xf_gpu = gpuarray.empty(self.shapeK, np.complex64)
        skfft.fft(x_gpu, xf_gpu, self.fftplan, False)
        return xf_gpu.get()

    def ifft(self, ff_fft):
        xf_gpu = gpuarray.to_gpu(ff_fft)
        x_gpu = gpuarray.empty(self.shapeX, np.float32)
        skfft.ifft(xf_gpu, x_gpu, self.ifftplan, False)
        return x_gpu.get()

    def compute_energy_from_Fourier(self, ff_fft):
        return (
            np.sum(abs(ff_fft[:, 0]) ** 2 + abs(ff_fft[:, -1]) ** 2)
            + 2 * np.sum(abs(ff_fft[:, 1:-1]) ** 2)
        ) / 2

    def compute_energy_from_spatial(self, ff):
        return np.mean(abs(ff) ** 2) / 2

    def project_fft_on_realX(self, ff_fft):
        return self.fft(self.ifft(ff_fft))


class FFTW2DReal2Complex:
    """A class to use fftw with float32.

    These ffts are NOT normalized (faster)!

    """

    type_real = "float32"
    type_complex = "complex64"

    def __init__(self, nx, ny):

        shapeX = [ny, nx]
        shapeK = [ny, nx // 2 + 1]

        self.shapeX = shapeX
        self.shapeK = shapeK

        self.arrayX = pyfftw.empty_aligned(shapeX, self.type_real)
        self.arrayK = pyfftw.empty_aligned(shapeK, self.type_complex)

        self.fftplan = pyfftw.FFTW(
            input_array=self.arrayX,
            output_array=self.arrayK,
            axes=(0, 1),
            direction="FFTW_FORWARD",
            threads=nthreads,
        )
        self.ifftplan = pyfftw.FFTW(
            input_array=self.arrayK,
            output_array=self.arrayX,
            axes=(0, 1),
            direction="FFTW_BACKWARD",
            threads=nthreads,
        )

        self.coef_norm = nx * ny

    def fft(self, ff):
        self.arrayX[:] = ff
        self.fftplan()
        return self.arrayK.copy()

    def ifft(self, ff_fft):
        self.arrayK[:] = ff_fft
        self.ifftplan(normalise_idft=False)
        return self.arrayX.copy()

    def compute_energy_from_Fourier(self, ff_fft):
        return (
            np.sum(abs(ff_fft[:, 0]) ** 2 + abs(ff_fft[:, -1]) ** 2)
            + 2 * np.sum(abs(ff_fft[:, 1:-1]) ** 2)
        ) / 2

    def compute_energy_from_spatial(self, ff):
        return np.mean(abs(ff) ** 2) / 2

    def project_fft_on_realX(self, ff_fft):
        return self.fft(self.ifft(ff_fft))


class FFTW2DReal2ComplexFloat64(FFTW2DReal2Complex):
    """A class to use fftw with float64."""

    type_real = "float64"
    type_complex = "complex128"



"""Servers for exec_async_servers (:mod:`fluidimage.executors.servers`)
=======================================================================

.. autofunction:: launch_server

.. autoclass:: Worker
   :members:
   :private-members:

.. autoclass:: WorkerMultiprocessing
   :members:
   :private-members:

.. autoclass:: WorkerServer
   :members:
   :private-members:

.. autoclass:: WorkerServerMultiprocessing
   :members:
   :private-members:

"""

import signal
import sys
import time
from multiprocessing import Event, Pipe, Process
from threading import Thread

import trio

from fluiddyn.io.tee import MultiFile
from fluidimage.util import cstring, log_debug, log_memory_usage, logger


def launch_server(
    topology,
    log_path,
    type_server="multiprocessing",
    sleep_time=0.1,
    logging_level="info",
):
    """Launch a server and return its client object"""

    parent_conn, child_conn = Pipe()

    event_has_to_stop = Event()

    # for testing
    if type_server == "multiprocessing":
        Process_ = Process
    elif type_server == "threading":
        Process_ = Thread
    else:
        raise ValueError

    in_process = Process_ == Process

    process = Process_(
        target=WorkerServerMultiprocessing,
        args=(
            child_conn,
            event_has_to_stop,
            type(topology),
            topology.params,
            sleep_time,
            in_process,
            log_path,
            logging_level,
        ),
    )
    process.daemon = True

    process.start()
    worker = WorkerMultiprocessing(parent_conn, event_has_to_stop, process)

    return worker


class Worker:
    def __init__(self, conn, event_has_to_stop, process):
        self.conn = conn
        self.event_has_to_stop = event_has_to_stop
        self.process = process
        self.nb_items_to_process = 0
        self.is_available = True
        self.is_unoccupied = True

    def well_done_thanks(self):
        self.nb_items_to_process -= 1
        if self.nb_items_to_process == 0:
            self.is_unoccupied = True

    def terminate(self):
        raise NotImplementedError


class WorkerMultiprocessing(Worker):
    def send_job(self, obj):
        self.is_unoccupied = False
        self.nb_items_to_process += 1
        self.send(obj)

    def send(self, obj):
        self.conn.send(obj)

    def new_pipe(self):
        return Pipe()

    def terminate(self):
        # self.conn.send("__terminate__")
        self.event_has_to_stop.set()
        if hasattr(self.process, "terminate"):
            self.process.terminate()


class WorkerServer:
    def __init__(self, sleep_time=0.01):
        self.sleep_time = sleep_time
        self.to_be_processed = []
        self.to_be_resent = []
        self._has_to_continue = True

        trio.run(self._start_async)

        # to avoid a pylint warning
        self.nursery = None
        self.t_start = None

    async def _start_async(self):
        async with trio.open_nursery() as self.nursery:
            self.nursery.start_soon(self.check_event_has_to_stop)
            self.nursery.start_soon(self.receive)
            self.nursery.start_soon(self.launch_works)
            self.nursery.start_soon(self.send)

    async def check_event_has_to_stop(self):
        raise NotImplementedError

    async def receive(self):
        raise NotImplementedError

    async def send(self):
        raise NotImplementedError

    async def launch_works(self):
        raise NotImplementedError


class WorkerServerMultiprocessing(WorkerServer):
    def __init__(
        self,
        conn,
        event_has_to_stop,
        topology_cls,
        params,
        sleep_time,
        in_process,
        log_path,
        logging_level,
    ):

        if in_process:

            def signal_handler(sig, frame):
                del sig, frame
                self._has_to_continue = False
                try:
                    self.nursery.cancel_scope.cancel()
                except AttributeError:
                    pass

            signal.signal(signal.SIGINT, signal_handler)

        self.conn = conn
        self.event_has_to_stop = event_has_to_stop
        self.topology = topology_cls(params)

        self._log_file = open(log_path, "w")

        stdout = sys.stdout
        if isinstance(stdout, MultiFile):
            stdout = sys.__stdout__

        stderr = sys.stderr
        if isinstance(stderr, MultiFile):
            stderr = sys.__stderr__

        sys.stdout = MultiFile([stdout, self._log_file])
        sys.stderr = MultiFile([stderr, self._log_file])

        if logging_level:
            for handler in logger.handlers:
                logger.removeHandler(handler)

            from fluidimage import config_logging

            config_logging(logging_level, file=sys.stdout)

        super().__init__(sleep_time=sleep_time)

    async def check_event_has_to_stop(self):
        while self._has_to_continue:
            if self.event_has_to_stop.is_set():
                self._has_to_continue = False
                break
            await trio.sleep(self.sleep_time)

    async def receive(self):
        while self._has_to_continue:
            ret = await trio.to_thread.run_sync(self.conn.recv)
            log_debug(f"receive: {ret}")
            if isinstance(ret, tuple) and ret[0] == "__t_start__":
                self.t_start = ret[1]
            else:
                self.to_be_processed.append(ret)

    async def launch_works(self):
        while self._has_to_continue:
            while not self.to_be_processed:
                await trio.sleep(self.sleep_time)

            work_name, key, obj, child_conn = self.to_be_processed.pop(0)
            work = self.topology.works_dict[work_name]

            def do_the_job(work, obj):
                return work.func_or_cls(obj)

            t_start = time.time()

            log_memory_usage(
                f"{time.time() - self.t_start:.2f} s. Launch work "
                + work.name_no_space
                + f" ({key}). mem usage"
            )
            # pylint: disable=W0703
            try:
                result = await trio.to_thread.run_sync(do_the_job, work, obj)
            except Exception as error:
                logger.error(
                    cstring(
                        "error during work " f"{work.name_no_space} ({key})",
                        color="FAIL",
                    )
                )
                result = error
            else:
                logger.info(
                    f"work {work.name_no_space} ({key}) "
                    f"done in {time.time() - t_start:.3f} s"
                )

            self.to_be_resent.append((work_name, key, result, child_conn))

    async def send(self):
        while self._has_to_continue:
            while not self.to_be_resent:
                await trio.sleep(self.sleep_time)
            work_name, key, result, child_conn = self.to_be_resent.pop(0)

            log_debug(f"send {work_name}, {key}")
            await trio.to_thread.run_sync(
                child_conn.send, (work_name, key, result)
            )



import pytest

# just to be sure that matplotlib is importable
import matplotlib

try:
    import matplotlib.backends.qt_compat
except ImportError:
    import_error_qt_compat = True
else:
    import_error_qt_compat = False

# For Travis where PyQt5 can't be installed (why?)
@pytest.mark.skipif(
    import_error_qt_compat, reason="ImportError matplotlib.backends.qt_compat"
)
def test_launcher(qtbot):
    from .main import Program

    widget = Program()
    qtbot.addWidget(widget)



import os
from configparser import ConfigParser


def get_config():
    config = ConfigParser()

    home = os.path.expanduser("~")
    path_config = os.path.join(home, ".fluidimagerc")

    if os.path.exists(path_config):
        config.read(path_config)

    config_dict = {}
    for section in config.sections():
        section_dict = {}
        for option in config.options(section):
            value = config.get(section, option)
            if value.lower in ["true", "false"]:
                value = value == "true"
            section_dict[option] = value
        config_dict[section] = section_dict

    return config_dict



import resource
from time import clock

import matplotlib.pyplot as plt
import numpy as np

from fluidimage.calcul.interpolate.thin_plate_spline import (
    ThinPlateSpline,
    compute_tps_coeff,
)


def get_memory_usage():
    return resource.getrusage(resource.RUSAGE_SELF).ru_maxrss / 1000.0


def bench(n0, n1):
    # set of random x coordinates from 0 to 2pi
    x = 2 * np.pi * np.random.rand(n0)
    y = 2 * np.pi * np.random.rand(n0)

    U = x

    # calculate tps coeff
    centers = np.vstack([x, y])

    t = clock()
    U_smooth, U_tps = compute_tps_coeff(centers, U, 0)
    t_tps_coeff = clock() - t

    # interpolation grid
    xI = yI = np.linspace(0, 2 * np.pi, n1)
    XI, YI = np.meshgrid(xI, yI)
    XI = XI.ravel()
    YI = YI.ravel()

    new_positions = np.vstack([XI, YI])

    # evaluate interpolation on the new grid
    t = clock()
    tps = ThinPlateSpline(new_positions, centers)
    memory = get_memory_usage()
    t_init = clock() - t

    t = clock()
    U_eval = tps.compute_field(U_tps)
    t_eval = clock() - t

    t = clock()
    DUX_eval, DUY_eval = tps.compute_gradient(U_tps)
    t_eval_grad = clock() - t

    return t_tps_coeff, t_init, t_eval, t_eval_grad, memory


ns = np.array([10, 20, 40, 60, 80, 100, 140, 200, 400, 600, 800, 1000])

t_tps_coeff = np.empty(ns.shape)
t_init = np.empty_like(t_tps_coeff)
t_eval = np.empty_like(t_tps_coeff)
t_eval_grad = np.empty_like(t_tps_coeff)
memory = np.empty_like(t_tps_coeff)

n0 = 20

for i, n1 in enumerate(ns):
    a, b, c, d, memory[i] = bench(n0, n1)
    t_tps_coeff[i] = a
    t_init[i] = b
    t_eval[i] = c
    t_eval_grad[i] = d


plt.figure()
ax = plt.gca()

ax.plot(ns, t_tps_coeff / ns, "b")
ax.plot(ns, t_init / ns, "r")
ax.plot(ns, t_eval / ns, "g")
ax.plot(ns, t_eval_grad / ns, "y")


plt.figure()
ax = plt.gca()

ax.plot(ns, memory, "r")

ax.set_ylabel("memory (Mo)")

plt.show()



import sys
import unittest

from fluiddyn.io import stdout_redirected
from fluidimage import get_path_image_samples
from fluidimage.gui.imviewer import ImageViewer, parse_args


path_image_samples = get_path_image_samples()


class TestImageViewer(unittest.TestCase):
    def test_main(self):
        with stdout_redirected():
            command = "fluidimviewer"
            args = command.split()
            args.append(str(path_image_samples / "Karman/Images"))
            sys.argv = args
            args = parse_args()
            ImageViewer(args)

            args = command.split()
            args.append(str(path_image_samples / "Karman/Images/*"))
            sys.argv = args

            args = parse_args()
            self = ImageViewer(args)

            self.set_autoclim(None)

            self._switch()
            self._switch()

            self._increase_ifile()
            self._decrease_ifile()

            self._submit_n("2")

            self._increase_ifile_n()
            self._decrease_ifile_n()

            self._change_cmin("1")
            self._change_cmax("2")



"""Parse and analyze logging files (:mod:`fluidimage.topologies.log`)
=====================================================================

.. autoclass:: LogTopology
   :members:
   :private-members:

"""

import time
from glob import glob
from pathlib import Path

import matplotlib.pyplot as plt
import numpy as np

from fluiddyn.util import is_run_from_ipython

if is_run_from_ipython():
    plt.ion()

colors = ["r", "b", "y", "g"]


def float_no_valueerror(word):
    try:
        return float(word)
    except ValueError:
        return np.nan


class LogTopology:
    """Parse and analyze logging files."""

    def __init__(self, path):
        path = Path(path)

        if path.is_dir():
            paths = sorted(glob(str(path / "log_*")))
            paths = [path for path in paths if "_multi" not in Path(path).name]
            if len(paths) == 0:
                raise ValueError("No log files found in the current directory.")

            path = Path(paths[-1])
            if path.is_dir():
                path = Path(
                    next(
                        path
                        for path in glob(str(path / "log*"))
                        if "_multi" not in path
                    )
                )

        self.log_dir_path = path.parent
        self.log_file = path.name
        self._title = str(self.log_file)

        self._parse_log(path)

    def _parse_log(self, path):
        self.works = works = []
        self.works_ended = works_ended = []
        self.nb_cpus_allowed = None
        self.nb_max_workers = None
        self.log_files = None
        self.executor_name = None
        self.topology_name = None
        with open(path, "r") as logfile:
            print("Parsing log file: ", path)
            for iline, line in enumerate(logfile):
                if iline % 100 == 0:
                    print(f"\rparse line {iline}", end="", flush=True)

                if line.startswith("ERROR: "):
                    continue

                if self.nb_cpus_allowed is None and line.startswith(
                    "INFO:   nb_cpus_allowed = "
                ):
                    self.nb_cpus_allowed = int(line.split()[3])
                    self._title += f", nb_cpus_allowed = {self.nb_cpus_allowed}"

                if self.nb_max_workers is None and line.startswith(
                    "INFO:   nb_max_workers = "
                ):
                    self.nb_max_workers = int(line.split()[3])
                    self._title += f", nb_max_workers = {self.nb_max_workers}"

                if self.topology_name is None:
                    begin = "INFO:   topology: "
                    if line.startswith(begin):
                        self.topology_name = line.split(begin)[1].strip()

                if self.executor_name is None:
                    begin = "INFO:   executor: "
                    if line.startswith(begin):
                        self.executor_name = line.split(begin)[1].strip()

                if self.log_files is None:
                    begin = "INFO: logging files: "
                    if line.startswith(begin):
                        self.log_files = eval(line.split(begin)[1].strip())

                if line.startswith("INFO: ") and ". mem usage: " in line:
                    line = line[11:]
                    words = line.split()

                    try:
                        mem = float(words[-2])
                    except ValueError:
                        pass

                    if ". Launch work " in line:
                        name = words[4]
                        key = words[5][1:-2]
                        t = float(words[0])
                        works.append(
                            {
                                "name": name,
                                "key": key,
                                "mem_start": mem,
                                "time": t,
                            }
                        )
                    else:
                        date = words[0][:-1]
                        t = time.mktime(
                            time.strptime(date[:-3], "%Y-%m-%d_%H-%M-%S")
                        ) + float(date[-3:])

                    if ": starting execution. mem usage" in line:
                        self.date_start = date
                        self.mem_start = mem
                        time_start = t
                    elif ": end of `compute`. mem usage" in line:
                        self.date_end = date
                        self.duration = t - time_start
                        self.mem_end = mem

                if line.startswith("INFO: work "):
                    words = line.split()
                    name = words[2]
                    key = words[3][1:-1]
                    try:
                        duration = float(words[-2])
                    except ValueError:
                        pass
                    else:
                        works_ended.append(
                            {"name": name, "key": key, "duration": duration}
                        )

        print("\rdone" + 20 * " ")

        if self.log_files is not None:
            path_dir = self.log_dir_path
            for file_name in self.log_files:
                path = path_dir / file_name
                with open(path, "r") as logfile:
                    print("Parsing log file: ", path.name)
                    for iline, line in enumerate(logfile):
                        if iline % 100 == 0:
                            print(f"\rparse line {iline}", end="", flush=True)

                        if line.startswith("ERROR: "):
                            continue

                        if line.startswith("INFO: ") and ". mem usage: " in line:
                            line = line[11:]
                            words = line.split()
                            mem = float_no_valueerror(words[-2])

                            if ". Launch work " in line:
                                name = words[4]
                                key = words[5][1:-2]
                                t = float_no_valueerror(words[0])
                                works.append(
                                    {
                                        "name": name,
                                        "key": key,
                                        "mem_start": mem,
                                        "time": t,
                                    }
                                )
                            else:
                                date = words[0][:-1]
                                t = time.mktime(
                                    time.strptime(date[:-3], "%Y-%m-%d_%H-%M-%S")
                                ) + float_no_valueerror(date[-3:])

                            if ": starting execution. mem usage" in line:
                                self.date_start = date
                                self.mem_start = mem
                                time_start = t
                            elif ": end of `compute`. mem usage" in line:
                                self.date_end = date
                                self.duration = t - time_start
                                self.mem_end = mem

                        if line.startswith("INFO: work "):
                            words = line.split()
                            name = words[2]
                            key = words[3][1:-1]
                            duration = float_no_valueerror(words[-2])
                            works_ended.append(
                                {"name": name, "key": key, "duration": duration}
                            )
                    print("\rdone" + 20 * " ")

        self.names_works = names_works = []
        for work in works:
            if work["name"] not in names_works:
                names_works.append(work["name"])

        self.durations = durations = {}
        self.times = times = {}
        self.keys = keys = {}
        for name in self.names_works:
            times[name] = []
            keys[name] = []
            for work in self.works:
                if work["name"] == name:
                    times[name].append(work["time"])
                    keys[name].append(work["key"])

            durations[name] = []
            for key in keys[name]:
                founded = False
                for work in self.works_ended:
                    if work["name"] == name and work["key"] == key:
                        durations[name].append(work["duration"])
                        founded = True
                        break

                if not founded:
                    durations[name].append(np.nan)

    def plot_memory(self):
        """Plot the memory usage versus time."""
        plt.figure()
        ax = plt.gca()
        ax.set_xlabel("time (s)")
        ax.set_ylabel("memory (Mo)")
        ax.set_title(self._title, fontdict={"fontsize": 12})

        memories = np.empty(len(self.works))
        times = np.empty(len(self.works))
        for i, work in enumerate(self.works):
            memories[i] = work["mem_start"]
            times[i] = work["time"]

        ax.plot(times, memories, "o-")
        ax.plot(0, self.mem_start, "x")
        if hasattr(self, "duration"):
            ax.plot(self.duration, self.mem_end, "x")
        plt.show()

    def plot_durations(self):
        """Plot the duration of the works."""
        plt.figure()
        ax = plt.gca()
        ax.set_xlabel("time (s)")
        ax.set_ylabel("duration (s)")
        ax.set_title(self._title, fontdict={"fontsize": 12})

        lines = []

        for i, name in enumerate(self.names_works):
            times = np.array(self.times[name])
            durations = self.durations[name]
            (l,) = ax.plot(times, durations, colors[i] + "o")
            lines.append(l)

            for it, t in enumerate(times):
                d = durations[it]
                ax.plot([t, t + d], [d, d], colors[i])

            d = np.nanmean(durations)
            ax.plot(
                [times.min(), times.max()], [d, d], colors[i] + ":", linewidth=2
            )

        ax.legend(lines, self.names_works, loc="center left", fontsize="x-small")

        plt.show()

    def plot_nb_workers(self, str_names=None):
        """Plot the number of workers versus time."""
        if str_names is not None:
            names = [name for name in self.names_works if str_names in name]
        else:
            names = self.names_works

        nb_workers = {}
        times = {}
        for name in names:
            times_start = list(self.times[name])
            times_stop = list(
                np.array(times_start) + np.array(self.durations[name])
            )

            deltas = np.array([1 for t in times_start] + [-1 for t in times_stop])
            times_unsorted = np.array(times_start + times_stop)

            argsort = np.argsort(times_unsorted)

            ts = times_unsorted[argsort]
            nbws = np.cumsum(deltas[argsort])

            ts2 = []
            nbws2 = []
            for i, t in enumerate(ts[:-1]):
                nbw = nbws[i]
                ts2.append(t)
                ts2.append(ts[i + 1])
                nbws2.append(nbw)
                nbws2.append(nbw)

            times[name] = ts2
            nb_workers[name] = nbws2

        plt.figure()
        ax = plt.gca()
        ax.set_xlabel("time (s)")
        ax.set_ylabel("number of workers")
        ax.set_title(self._title, fontdict={"fontsize": 12})

        lines = []

        for i, name in enumerate(self.names_works):
            (l,) = ax.plot(times[name], nb_workers[name], colors[i] + "-")
            lines.append(l)

        ax.legend(lines, names, loc="center left", fontsize="x-small")

        plt.show()



"""PIV post-processing (:mod:`fluidimage.postproc.piv`)
=======================================================

.. autofunction:: get_grid_pixel_from_piv_file

.. autofunction:: get_grid_pixel

.. autoclass:: PIV2d
   :members:
   :private-members:

.. autoclass:: ArrayPIV
   :members:
   :private-members:

"""


import h5py


from fluiddyn.util.paramcontainer import ParamContainer
from fluidimage.works.piv.multipass import WorkPIV
from .vector_field import VectorFieldOnGrid, ArrayOfVectorFieldsOnGrid


class PIV2d(VectorFieldOnGrid):
    pass


def get_grid_pixel_from_piv_file(path, index_pass=-1):
    """Recompute 1d arrays containing the approximate positions of the vectors

    Useful to compute a grid on which we can interpolate the displacement fields.

    Parameters
    ----------

    path: str

      Path of a PIV file.

    index_pass: int

      Index of the pass

    Returns
    -------

    xs1d: np.ndarray

      Indices (2nd, direction "x") of the pixel in the image

    ys1d: np.ndarray

      Indices (1st, direction "y") of the pixel in the image

    """
    with h5py.File(path, "r") as file:
        params = ParamContainer(hdf5_object=file["params"])
        shape_images = file["couple/shape_images"][...]

    return get_grid_pixel(params, shape_images, index_pass)


def get_grid_pixel(params, shape_images, index_pass=-1):
    """Recompute 1d arrays containing the approximate positions of the vectors

    Useful to compute a grid on which we can interpolate the displacement fields.

    Parameters
    ----------

    params: fluiddyn.util.paramcontainer.ParamContainer

      Parameters for the class :class:`fluidimage.works.piv.multipass.WorkPIV`

    shape_images: sequence

      Shape of the images

    index_pass: int

      Index of the pass

    Returns
    -------

    xs1d: np.ndarray

      Indices (2nd, direction "x") of the pixel in the image

    ys1d: np.ndarray

      Indices (1st, direction "y") of the pixel in the image


    """

    params_default = WorkPIV.create_default_params()
    params_default._modif_from_other_params(params)

    work_multi = WorkPIV(params_default)
    work_multi._prepare_with_image(imshape=shape_images)

    work = work_multi.works_piv[index_pass]
    xs1d = work.ixvecs
    ys1d = work.iyvecs

    return xs1d, ys1d


class ArrayPIV(ArrayOfVectorFieldsOnGrid):
    """Array of PIV fields on a regular grid."""



import numpy as np


def make_synthetic_images(
    displacements,
    nb_particles,
    shape_im0,
    shape_im1=None,
    epsilon=0.0,
    part_size=np.sqrt(1 / 0.5),
):
    ny, nx = tuple(shape_im0)

    displacement_x, displacement_y = tuple(displacements)

    xs = np.arange(nx, dtype="float32")
    ys = np.arange(ny, dtype="float32")

    Xs, Ys = np.meshgrid(xs, ys)

    xmax = xs.max() + abs(displacement_x)
    xmin = xs.min() - abs(displacement_x)
    xparts = [
        xmin + (xmax - xmin) * np.random.rand() for i in range(nb_particles)
    ]
    ymax = ys.max() + abs(displacement_y)
    ymin = ys.min() - abs(displacement_y)
    yparts = [
        ymin + (ymax - ymin) * np.random.rand() for i in range(nb_particles)
    ]

    def f(x, y):
        result = np.zeros_like(x, dtype="float32")

        for xpart, ypart in zip(xparts, yparts):
            result += np.exp(
                -(1 / part_size**2) * ((x - xpart) ** 2 + (y - ypart) ** 2)
            )

        return result

    im0 = f(Xs, Ys)
    im1 = f(Xs - displacement_x, Ys - displacement_y)

    if epsilon > 0:
        im0 += epsilon * np.random.randn(*im0.shape)
        im1 += epsilon * np.random.randn(*im1.shape)

    if shape_im1 is not None:
        ny1, nx1 = tuple(shape_im1)
        if ny1 > ny or nx1 > nx:
            raise ValueError("ny1 > ny or nx1 > nx")

        ixfirst = (nx - nx1) // 2
        iyfirst = (ny - ny1) // 2
        im1 = im1[iyfirst : iyfirst + ny1, ixfirst : ixfirst + nx1]

    return im0, im1



import unittest
from time import time

import numpy as np

from fluidimage.calcul.correl import CUFFT2DReal2Complex, FFTW2DReal2Complex

# from scipy.misc import lena
# from correl import calcul_correl_norm_scipy, CorrelWithFFT


class TestFFTW2DReal2Complex(unittest.TestCase):

    # def test_correl(self):

    #     rtime, ntime = 0., 0.

    #     im1 = lena()
    #     im2 = im1 + np.random.randn(*im1.shape) * 50  # add noise
    #     im1 = im1.astype('float32')
    #     im2 = im2.astype('float32')
    #     nx, ny = im1.shape

    #     t0 = time()
    #     correl = CorrelWithFFT(nx, ny)
    #     correlfft = correl.calcul_correl_norm(im1, im2)
    #     ntime += time() - t0

    #     t0 = time()
    #     correl_dir = calcul_correl_norm_scipy(im1, im2)
    #     rtime += time() - t0

    #     print 'correl fft speedup = %g' % (rtime / ntime)
    #     # ax1 = plt.subplot(121)
    #     # ax2 = plt.subplot(122)
    #     # ax1.imshow(correlfft)
    #     # ax2.imshow(correl_dir)
    #     # plt.show(block=True)

    #     rtol = 8e-0
    #     atol = 3e-04
    #     # tmp = np.absolute(correlfft - correl_dir) #- atol - rtol*np.abs(back)

    #     # print(tmp.max())
    #     self.assertTrue(np.allclose(correlfft,
    #                                 correl_dir, rtol=rtol, atol=atol))

    def test_fft(self):
        """simple"""
        nx = 4
        ny = 2
        op = FFTW2DReal2Complex(nx, ny)

        func_fft = np.zeros(op.shapeK, dtype=op.type_complex)
        func_fft[0, 1] = 1

        self.compute_and_check(func_fft, op)

    def compute_and_check(self, func_fft, op):

        energyK = op.compute_energy_from_Fourier(func_fft)

        func = op.ifft(func_fft)
        energyX = op.compute_energy_from_spatial(func)

        back_fft = op.fft(func) / op.coef_norm
        back = op.ifft(back_fft)

        rtol = 8e-05
        atol = 1e-04
        self.assertTrue(np.allclose(func_fft, back_fft, rtol=rtol, atol=atol))
        self.assertTrue(np.allclose(func, back, rtol=rtol, atol=atol))

        self.assertAlmostEqual(energyX / energyK, 1.0, places=3)

        energyKback = op.compute_energy_from_Fourier(back_fft)
        self.assertAlmostEqual(energyK / energyKback, 1.0, places=3)

    def compute_and_check2(self, func, op):

        energyX = op.compute_energy_from_spatial(func)
        func_fft = op.fft(func)
        energyK = op.compute_energy_from_Fourier(func_fft)
        back = op.ifft(func_fft)
        energyX = op.compute_energy_from_spatial(back)
        func_fft_2 = op.fft(back)
        energyKback = op.compute_energy_from_Fourier(func_fft_2)
        rtol = 8e-05
        atol = 1e-04
        self.assertTrue(np.allclose(func_fft, func_fft_2, rtol=rtol, atol=atol))

        # tmp = np.absolute(func - back) - atol - rtol*np.abs(back)

        # print(tmp.max())

        self.assertTrue(np.allclose(func, back, rtol=rtol, atol=atol))

        self.assertAlmostEqual(energyX / energyK, 1.0, places=3)
        self.assertAlmostEqual(energyK / energyKback, 1.0, places=3)

    def bench_fft_random(self):
        """random"""
        nx = 128 * 16
        ny = 64 * 4 * 8

        rtime, ntime = 0.0, 0.0
        Nloops = 2
        for nloop in range(Nloops):

            op = FFTW2DReal2Complex(nx, ny)

            func = np.random.random(op.shapeX)

            func = np.array(func, dtype=op.type_real)

            t0 = time()
            func_fft = op.fft(func)
            func = op.ifft(func_fft)

            self.compute_and_check2(func, op)
            ntime += time() - t0

            op = CUFFT2DReal2Complex(nx, ny)

            t0 = time()
            func_fft = op.fft(func)
            func1 = op.ifft(func_fft)

            self.compute_and_check2(func1, op)
            rtime += time() - t0

        print(
            "array size = %5d x %5d : gpu speedup = %g" % (nx, ny, ntime / rtime)
        )


if __name__ == "__main__":
    unittest.main()



"""Multipass PIV (:mod:`fluidimage.works.piv.multipass`)
========================================================

.. autoclass:: WorkPIV
   :members:
   :private-members:

"""

from copy import copy

from fluiddyn.util.paramcontainer import ParamContainer

from .fix import WorkFIX
from .singlepass import FirstWorkPIV, InterpError, WorkPIVFromDisplacement
from .. import BaseWork
from ...data_objects.piv import MultipassPIVResults


class WorkPIV(BaseWork):
    """Main work for PIV with multipass.

    Parameters
    ----------

    params : :class:`fluiddyn.util.paramcontainer.ParamContainer`

      The default parameters are obtained from the class method
      :func:`WorkPIV.create_default_params`.

    Notes
    -----

    Steps for a computation:

    - first estimation of the PIV field with a work
      :class:`fluidimage.works.piv.singlepass.FirstWorkPIV`

    - fix (remove "false vectors") with a work
      :class:`fluidimage.works.piv.fix.WorkFIX`

    - depending of the value of `params.multipass.number`, iteration to compute
      PIV displacements with the works
      :class:`fluidimage.works.piv.singlepass.WorkPIVFromDisplacement` and
      :class:`fluidimage.works.piv.fix.WorkFIX`.

    """

    @classmethod
    def create_default_params(cls):
        "Create an object containing the default parameters (class method)."
        params = ParamContainer(tag="params")
        cls._complete_params_with_default(params)
        return params

    @classmethod
    def _complete_params_with_default(cls, params):
        """Complete the default parameters (class method)."""
        FirstWorkPIV._complete_params_with_default(params)
        WorkFIX._complete_params_with_default(params)

        params._set_child(
            "multipass",
            attribs={
                "number": 1,
                "coeff_zoom": 2,
                "use_tps": "last",
                "subdom_size": 200,
                "smoothing_coef": 0.5,
                "threshold_tps": 1.0,
            },
        )

        params.multipass._set_doc(
            """Multipass PIV parameters:

number : int (default 1)
    Number of PIV passes.

coeff_zoom : integer or iterable of size `number - 1`.

    Reduction coefficient defining the size of the interrogation windows for
    the passes 1 (second pass) to `number - 1` (last pass) (always
    defined comparing the passes `i-1`).

use_tps : bool or 'last'

    If it is True, the interpolation is done using the Thin Plate Spline method
    (computationnally heavy but sometimes interesting). If it is 'last', the
    TPS method is used only for the last pass.

subdom_size : int

    Number of vectors in the subdomains used for the TPS method.

smoothing_coef : float

    Coefficient used for the TPS method. The result is smoother for larger
    smoothing_coef.

threshold_tps :  float

    Allowed difference of displacement (in pixels) between smoothed and input
    field for TPS filter.

"""
        )

    def __init__(self, params=None):

        self.params = params

        self.works_piv = []
        self.works_fix = []

        work_piv = FirstWorkPIV(params)
        self.works_piv.append(work_piv)
        self.works_fix.append(WorkFIX(params.fix, work_piv))

        coeff_zoom = params.multipass.coeff_zoom

        if isinstance(coeff_zoom, int):
            coeffs_zoom = [coeff_zoom] * (params.multipass.number - 1)
        elif len(coeff_zoom) == params.multipass.number - 1:
            coeffs_zoom = coeff_zoom
        else:
            raise ValueError(
                "params.multipass.coeff_zoom has to be an integer or "
                "an iterable of length params.multipass.number - 1"
            )

        shape_crop_im0 = copy(params.piv0.shape_crop_im0)
        shape_crop_im1 = copy(params.piv0.shape_crop_im1)
        if shape_crop_im1 is None:
            shape_crop_im1 = shape_crop_im0

        if isinstance(shape_crop_im0, int):
            shape_crop_im0 = (shape_crop_im0, shape_crop_im0)
        elif (
            isinstance(shape_crop_im0, (list, tuple)) and len(shape_crop_im0) == 2
        ):
            shape_crop_im0 = tuple(shape_crop_im0)
        else:
            raise NotImplementedError(
                "For now, shape_crop_im0 has to be one or two integer!"
            )

        if isinstance(shape_crop_im1, int):
            shape_crop_im1 = (shape_crop_im1, shape_crop_im1)
        elif (
            isinstance(shape_crop_im1, (list, tuple)) and len(shape_crop_im1) == 2
        ):
            shape_crop_im1 = tuple(shape_crop_im1)
        else:
            raise NotImplementedError(
                "For now, shape_crop_im1 has to be one or two integer!"
            )

        for i in range(1, params.multipass.number):
            shape_crop_im0 = (
                copy(shape_crop_im0[0] / coeffs_zoom[i - 1]),
                copy(shape_crop_im0[1] / coeffs_zoom[i - 1]),
            )
            shape_crop_im1 = (
                copy(shape_crop_im1[0] / coeffs_zoom[i - 1]),
                copy(shape_crop_im1[1] / coeffs_zoom[i - 1]),
            )

            work_piv = WorkPIVFromDisplacement(
                params,
                index_pass=i,
                shape_crop_im0=shape_crop_im0,
                shape_crop_im1=shape_crop_im1,
            )
            self.works_piv.append(work_piv)
            self.works_fix.append(WorkFIX(params.fix, work_piv))

    def calcul(self, couple):
        """Compute a PIV field (multipass) from a couple of image."""

        results = MultipassPIVResults()

        # just for simplicity
        piv_result = couple

        for i, work_piv in enumerate(self.works_piv):
            work_fix = self.works_fix[i]
            piv_result = work_piv.calcul(piv_result)
            piv_result = work_fix.calcul(piv_result)
            results.append(piv_result)

        try:
            work_piv.apply_interp(piv_result, last=True)
        except InterpError as e:
            print("Warning: InterpError at the end of the last piv pass:", e)

        return results

    def _prepare_with_image(self, im=None, imshape=None):
        """Prepare the works PIV with an image."""
        if imshape is None:
            imshape = im.shape

        for work_piv in self.works_piv:
            work_piv._prepare_with_image(imshape=imshape)


params = WorkPIV.create_default_params()
__doc__ += params._get_formatted_docs()
